;+
; NAME:
;      ACS_TRAP_DENSITY
;
; CATEGORY:
;      Pixel-by-pixel correction of trailing due to charge transfer inefficiency in CCD
;      detectors.
;
; PURPOSE:
;      Model the density of charge traps (defects in the silicon substrate) in the HST
;      Advanced Camera for Surveys. The density grows over time.
;
; INPUTS:
;      DATES - Either a single floating point number or an array, which is the Modified
;              Julian Date(s) of interest.
;
; OPTIONAL INPUTS:
;      PARAMS- Shorthand alternative notation to enter following three parameters.
;      INITIA- Initial model density of charge traps per pixel on launch.
;      PERMAN- Rate of increase of traps per pixel per day that are permanent.
;      REMOVA- Rate of increase of traps per pixel per day that get fixed at next anneal.
;      ANNEAL- List of (MJD) dates when CCD was annealed DEFAULT: read in from file.
;      LAUNCH- Launch (MJD) date of camera DEFAULT: 2452334.5
;
; KEYWORD PARAMETERS:
;      None.
;
; OUTPUTS:
;      The density of charge traps per pixel on each date.
;      For the default model, this is the total density of traps; the density of traps with
;      release time tau=10.4 pixels is 3/4 that, and the density of traps with tau=0.88 pixels 
;      is that 1/4 that.
;
; MODIFICATION HISTORY:
;      Jul 09 - Need to read in an external file removed by RM, since no effect of anneals.
;      Jul 08 - Written by Richard Massey.
;-

; Read in a list of the dates (MJD) when ACS was annealed to remove charge traps.
function acs_find_anneals
;anneal_file="~/idl/cte/ACS_anneal_dates.txt"
;if not file_test(anneal_file) then message,"Cannot find list of anneal dates "+anneal_file
;readcol,anneal_file,format="A,I,I,I,I,I",month_string,day,year,hour,minute,second,/SILENT
;n_dates=n_elements(month_string) & if n_dates le 1 then message,"File format not recognised for list of anneal dates "+anneal_file
;month=intarr(n_dates) & for i=0,n_dates-1 do month[i]=where(strmatch(["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"],month_string[i]))+1
year=[2003,2003,2003,2003,2004,2004,2004,2004,2004,2004,2004,2004,2004,2004,2004,2004,2004,2004,2005,2005,2005,2005,2005,2005,2004,2005,2005,2005,2005,2005,2005,2005,2005]
month=[9,10,11,12,1,1,3,3,4,5,6,7,8,9,10,11,12,12,1,3,3,4,5,2,7,6,7,8,9,10,11,11,12]
day=[9,12,6,1,4,30,4,27,25,22,18,14,10,8,8,5,2,30,29,4,24,19,20,12,6,12,16,11,9,8,3,25,31]
hour=[8,7,15,15,9,1,1,0,21,7,4,16,4,6,9,19,14,12,11,22,9,7,5,9,16,22,7,20,8,10,15,8,4]
minute=[34,9,11,48,1,8,21,59,31,34,36,52,41,1,5,42,11,40,0,55,12,33,0,4,28,35,58,37,45,6,20,37,30]
second=[58,36,6,13,30,26,4,38,0,57,37,5,59,41,45,53,50,0,0,30,24,57,0,52,51,22,20,49,56,47,0,56,0]
jdcnv,year,month,day,hour+minute/60.+second/360.,dates
return, dates
end

; ************************************************************************

function acs_trap_density, DATES, PARAMS ,$
                           TRAP_INITIAL_DENSITY=trap_initial_density, $
                           TRAP_PERMANENT_RATE=trap_permanent_rate, $
                           TRAP_REMOVABLE_RATE=trap_removable_rate, $
                           ANNEAL_DATES=anneal_dates, $
                           LAUNCH_DATE=launch_date

COMPILE_OPT idl2

; Parse inputs
if keyword_set(params) then begin
  trap_initial_density=params[0]
  trap_permanent_rate=params[1]
  trap_removable_rate=params[2]
endif
if not keyword_set(launch_date)          then launch_date=2452334.5
if not keyword_set(trap_initial_density) then trap_initial_density=-0.176532 ; Illingworth
if not keyword_set(trap_permanent_rate)  then trap_permanent_rate=0.000455022
if not keyword_set(trap_removable_rate)  then trap_removable_rate=0
if not keyword_set(anneal_dates)         then if trap_removable_rate gt 0 then anneal_dates=acs_find_anneals() else anneal_dates=launch_date
;COSMOS 0.036513 4.3502d-4
;Illingworth -0.218267  0.000470158
;All post-SM4 -0.176532  0.000455022
;All, addititive drop 0.0246495 (-0.184300)  0.000448333 
; All, multiplicative drop 0.0127904 (*1-0.164000) 0.000459184

print,"*******",trap_initial_density

; Calculate overall linear growth of permanent traps
trap_density=trap_initial_density+(dates-launch_date)*trap_permanent_rate

; Add in those traps that will be fixed at the next annealing
if trap_removable_rate ne 0 then begin
  dates_to_adjust=where(dates gt min(anneal_dates) and dates lt max(anneal_dates), n_dates_to_adjust)
  if n_dates_to_adjust gt 0 then begin
    for i=0,n_dates_to_adjust-1 do begin
      date=dates[dates_to_adjust[i]]
      days_since_anneal=min(date-anneal_dates[where(date ge anneal_dates)])
      trap_density[dates_to_adjust[i]]=trap_density[dates_to_adjust[i]]+days_since_anneal*trap_removable_rate
    endfor
  endif
endif

; Tell the world
return,trap_density

end
