pro acs_find_warm_pixels, files, $
                          plots=plots, $
                          pixel_files=pixel_files, $
                          use_bad_columns=use_bad_columns, $
                          bad_column_factor=bad_column_factor, $
                          bad_column_loops=bad_column_loops, $
                          smooth_width=smooth_width, $
                          unsharp_masking_factor=unsharp_masking_factor, $
                          backgrounds=backgrounds

;+
; NAME:
;      ACS_FIND_WARM_PIXELS
;
; CATEGORY:
;      Pixel-by-pixel correction of trailing due to charge transfer inefficiency in CCD detectors.
;
; PURPOSE:
;      Identify useful-looking hot/warm pixels in an image and record the 
;      surrounding pixel values.
;
; INPUTS:
;      FILES                 - String array containing absolute file names.
;
; OPTIONAL INPUTS:
;      SMOOTH_WIDTH          - Width of (square) smoothing kernel for unsharp masking.
;      UNSHARP_MASKING_FACTOR- Factor multiplying the unfocussed version of the image during unsharp masking.
;      BAD_COLUMN_FACTOR     - Amount of sigma-clipping during the identification of bad columns.
;      BAD_COLUMN_LOOPS      - Number of iterations to perform during sigma-clipping of bad columns.
;
; KEYWORD PARAMETERS:
;      USE_BAD_COLUMNS       - Also look along columns with a serious defect? Default behaviour is to identify these then ignore them.
;      PLOTS                 - Draw some diagnostic plots to the screen?
;
; OUTPUTS:
;      PIXELS                - Optionally returns the structure as well as writing it to disc.
;      BACKGROUNDS           - A list of the background levels, in electrons.
;
; MODIFICATION HISTORY:
;      Feb 10 - Output of deduced background levels added by RM.
;      Jan 08 - Written by Richard Massey.
;-


; Parse inputs
n_files=n_elements(files)
if not keyword_set(bad_column_factor)      then bad_column_factor=3.5
if not keyword_set(bad_column_loops)       then bad_column_loops=5
if not keyword_set(smooth_width)           then smooth_width=3
if not keyword_set(unsharp_masking_factor) then unsharp_masking_factor=6;4
if not keyword_set(trail_length)           then trail_length=9
if not keyword_set(backgrounds)            then backgrounds=fltarr(n_files)

; Look at each image in turn
for i=0,n_files-1 do begin

  ; Read in file
  message,/info,"Reading image "+files[i]+".fits ("+strtrim(round(i*100./n_files),2)+"%)"
  fits_read,files[i]+".fits",image,header
  if strupcase(strtrim(sxpar(header,"BUNIT"),2)) ne "ELECTRONS" then message,"Image units not in electrons!"
  n_pixels=size(image,/DIMENSIONS)
  date=sxpar(header,"DATE")
 
  ; Identify bad columns, backround level and warm pixels
  if keyword_set(pixel_files) then begin
    
    ; Read in preexisting data
    message,/info,"Reading warm pixel locations from file "+pixel_files[i]+"_pixels.sav"
    restore,pixel_files[i]+"_pixels.sav"
    n_warm_pixels=n_elements(pixels) & warm_pixels_locations=[[pixels.x],[pixels.y]] & background=pixels(0).background
  endif else begin

    ; Work out from scratch. First, find bad columns (due to a really bright object or a pixel sufficiently hot to wipe everything above it)
    columns=total(image,2)/float(n_pixels[1])
    image_bad=bytarr(n_pixels[0],n_pixels[1])
    bad_columns=0
    if not keyword_set(use_bad_columns) then begin
      good_columns=indgen(n_pixels[0])
      for k=1,bad_column_loops do begin 
        mean_columns=median(columns[good_columns])
        stddev_columns=stddev(columns[good_columns])
        temp=where(abs(columns[good_columns]-mean_columns) ge bad_column_factor*stddev_columns,n_temp) & if n_temp gt 0 then bad_columns=[bad_columns,good_columns[temp]]
        good_columns=good_columns[where(abs(columns[good_columns]-mean_columns) lt bad_column_factor*stddev_columns)]
      endfor
      n_bad_columns=n_elements(bad_columns)-1
      if n_bad_columns gt 0 then begin
        bad_columns=bad_columns[1:*]
        for j=0,n_bad_columns-1 do image_bad[bad_columns[j],*]=1B
      endif
      message,/info,"Ignoring "+strtrim(n_bad_columns,2)+" bad columns"
    endif else n_bad_columns=0

    ; Subtract background level
    good_columns=where(total(float(image_bad),2) eq 0,n_good_columns)
    background=2.5*median(columns[good_columns])-1.5*mean(columns[good_columns])
    image_sub=image-background

    ; Unsharp mask image
    image_smo=smooth(image_sub,smooth_width,/edge_truncate)
    image_uns=image_sub-6*image_smo
    
    ; Ignore overscan regions, and very top and bottom of CCD since we can't follow trails that fall off the edges
    message,/info,"Ignoring very top and bottom of CCD"
    image_bad[*,0:trail_length-1]=1B
    image_bad[*,2048-1-trail_length:n_pixels[1]-1]=1B
    image_bad[0:n_pixels[0]-1-2048,*]=1B

    ; Identify hot pixels
    if not keyword_set(flux_range) then flux_range=[min(image,max=max),max]
    image_max=shift(image_sub,0,-1)>shift(image_sub,0,-2)>shift(image_sub,0,-3)>shift(image_sub,0,-4)>shift(image_sub,0,-5)>shift(image_sub,0,-6)>shift(image_sub,0,-7)>shift(image_sub,0,-8)>shift(image_sub,0,-9) $
             >shift(image_sub,0,1) >shift(image_sub,0,2) >shift(image_sub,0,3) >shift(image_sub,0,4) >shift(image_sub,0,5) >shift(image_sub,0,6) >shift(image_sub,0,7) >shift(image_sub,0,8) >shift(image_sub,0,9)
    warm_pixels=where($;image ge flux_range[0] and image lt flux_range[1] and $
                     image_bad eq 0B and $
                     image_sub gt image_max               and $ ; local maximum
                     image_sub gt shift(image_sub,1,0)    and $
                     image_sub gt shift(image_sub,-1,0)   and $
                     image_sub gt unsharp_masking_factor*shift(image_smo,1,0)  and $ ; even after unsharp masking
                     image_sub gt unsharp_masking_factor*shift(image_smo,-1,0) and $
                     image_sub gt unsharp_masking_factor*shift(image_smo,0,1)  and $
                     image_sub gt unsharp_masking_factor*shift(image_smo,0,-1),n_warm_pixels) 
    if n_warm_pixels eq 0 then message,"No pixels found!" & if n_warm_pixels eq 1 then message,"Only one pixel found!"
    warm_pixels_locations=[[warm_pixels mod n_pixels[0]],[warm_pixels/(n_pixels[0])]]
    
    ; Obtain date of exposure
    date=float(sxpar(header,"DATE"))

    ; Display results
    if keyword_set(plots) then begin
      inspect_frame=[1700,2000,1850,2040]
      inspect_frame=[0,300,1870,2057]
      display_spacing=10.
      window,0,title="Background subtracted image"
      usersym, cos(2*!pi*findgen(21)/20), sin(2*!pi*findgen(21)/20) ; Set psym=8 to be a circle
      loadct,0,/silent
      plt_image,image_sub[inspect_frame[0]:inspect_frame[1]-1,inspect_frame[2]:inspect_frame[3]-1],/iso,frame=inspect_frame,cran=[-1,1]*background[0]*0.1
      loadct,4,/silent
      oplot,psym=6,col=150,warm_pixels_locations[*,0]+0.5,warm_pixels_locations[*,1]+0.5
      for j=0,n_bad_columns-1 do oplot,psym=7,symsize=2,col=50,replicate(bad_columns[j],(n_pixels[1]+2)/float(display_spacing))+0.5,findgen((n_pixels[1]+2)/float(display_spacing)+randomn(j+12,1)/100.)*display_spacing
      loadct,0,/silent
      window,1,title="Unsharp masked image"
      plt_image,image_uns[inspect_frame[0]:inspect_frame[1]-1,inspect_frame[2]:inspect_frame[3]-1],/iso,frame=inspect_frame,cran=[-1,1]*background[0]*0.1
      loadct,4,/silent
      oplot,psym=6,col=150,warm_pixels_locations[*,0]+0.5,warm_pixels_locations[*,1]+0.5
      for j=0,n_bad_columns-1 do oplot,psym=7,symsize=2,col=50,replicate(bad_columns[j],(n_pixels[1]+2)/float(display_spacing))+0.5,findgen((n_pixels[1]+2)/float(display_spacing)+randomu(seed,1)/10.)*display_spacing
      window,2,title="Image cross section"
      loadct,0,/silent
      plot,columns,psym=-3,/xstyle,thick=1,/ynozero,/nodata
      loadct,4,/silent
      oplot,columns,psym=-3,thick=1,col=50
      loadct,0,/silent
      columns_temp=columns & columns_temp[bad_columns]=!values.f_nan
      oplot,columns_temp,psym=-3,thick=1
     ;for j=0,n_bad_columns-1 do oplot,psym=-3,col=50,thick=1,replicate(bad_columns[j],2),[-1e6<min(columns,max=max),max>1e6]*2
    endif

  endelse
  message,/info,"Found "+strtrim(n_warm_pixels,2)+" warm pixels. Background level is "+strtrim(background,2)+" electrons."
  
;  ; Save pixel locations to disc as a DS9 region file
;  message,/info,"Writing DS9 regions to file "+files[i]+"_pixels.reg"
;  openw,lun,files[i]+"_pixels.reg",/get_lun
;  printf,lun,"# Region file format: DS9 version 4.0"
;  printf,lun,"# Filename: "+files[i]+".fits"
;  printf,lun,"global color=red font="+'"'+"helvetica 10 normal"+'"'+" select=1 highlite=1 edit=1 move=1 delete=1 include=1 fixed=0 source"
;  printf,lun,"physical"
;  for j=0L,n_warm_pixels-1 do printf,lun,"circle("+strtrim(warm_pixels_locations[j,0]+1,2)+","+strtrim(warm_pixels_locations[j,1]+1,2)+",1)"
;  close,lun
;  free_lun,lun

  ; Save all findings to disc as an IDL binary save file
  backgrounds[i]=background
  pixels=replicate({pixel,x:0U,y:0U,flux:0L,background:0.,date:0.,leading:lonarr(9),trailing:lonarr(9)},n_warm_pixels)
  for j=0L,n_warm_pixels-1 do begin
    pixels[j]={pixel,x:warm_pixels_locations[j,0],y:warm_pixels_locations[j,1],$
                     flux:image[warm_pixels_locations[j,0],warm_pixels_locations[j,1]],background:background,date:date,$
                     leading:reverse(reform(image[warm_pixels_locations[j,0],warm_pixels_locations[j,1]-9:warm_pixels_locations[j,1]-1])),$
                     trailing:reform(image[warm_pixels_locations[j,0],warm_pixels_locations[j,1]+1:warm_pixels_locations[j,1]+9])}
  endfor
  message,/info,"Saving data to file "+files[i]+"_pixels.sav"
  save,pixels,file=files[i]+"_pixels.sav"
  print

endfor

end
