function acs_cte_model, x, params

; NAME:
;      ACS_CTE_MODEL
;
; CATEGORY:
;      Pixel-by-pixel correction of trailing due to charge transfer inefficiency in CCD detectors.
;
; PURPOSE:
;      A model for a CTE model as a function of date, pixel flux, CCD position etc.
;
; INPUTS:
;      X      - An Nx4 floating point array, where the 4 are [N_TRANS,FLUX,DATE,BACKGROUND]
;      PARAMS - A 3, 4 or 5-element floating point array (one more than the number of variables in X).
;               The first parameter is an overall normalisation (i.e. proportional to mean trap density)
;               The second parameter is the power to which the number of transfers is raised
;               The third parameter is the power to which the flux is raised
;               The fourth parameter is the effective notch depth
;               The fifth parameter is the growth rate of traps per day; it is not used (and should not be passed) if MJD is not supplied 
;               The sixth parameter is the power to which the background level is raised; it is not used (and should not be passed) if background level is not supplied 
;
; OPTIONAL INPUTS:
;      None.
;
; KEYWORD PARAMETERS:
;      
;
; OUTPUTS:
;      
;
; MODIFICATION HISTORY:
;      Jan 08 - Written by Richard Massey.

; Extract variables into nicer local names
density=params[0]
growth_rate=params[1]
date_power=params[2]
tran_power=params[3]
n_pixels=params[4]
notch_depth=params[5]
full_well_depth=params[6]
flux_power=params[7]
back_power=params[8]

; Extract ordinates
tran=x[*,0]
flux=x[*,1]
date=x[*,2]
back=x[*,3]

; Compute model
return, (density+growth_rate*date^date_power)$
       *(   (((flux-notch_depth)>0)/float(full_well_depth-notch_depth))^flux_power*(tran)^tran_power $
           -(((back-notch_depth)>0)/float(full_well_depth-notch_depth))^back_power                                      )

end

;************************************************************


function acs_cte_trail_model, x, params, delta_params

; NAME:
;      ACS_FIT_TRAIL_MODEL
;
; PURPOSE:
;      A multiple exponential model for a CTE trail.
;
; INPUTS:
;      X      -
;      PARAMS - floating point vector with an even number of elements that
;               alternately contain an amplitude and an e-folding factor of
;               one exponential.
;
; MODIFICATION HISTORY:
;      Jan 08 - Written by Richard Massey.

; Check input parameters are acceptable
if n_elements(params) eq 0 then message,"Must specify some model parameters!"
;if min(params) lt 0 then print,"WARNING: negative parameter ",params

; Compute model using these parameters
model=replicate(x[0]*0.,n_elements(x))
for i=0,n_elements(params)-(n_elements(params) mod 2)-1,2 do model=model+params[i]*exp(-x/float(params[i+1]))/total(exp((-0.5-indgen(10*params[i+1]))/float(params[i+1])))
if (n_elements(params) mod 2) eq 1 then model=model+(reverse(params))[0]

;; Compute analytic derivatives
;if n_params() GT 2 then begin
;  delta_params = replicate(x[0],n_elements(x),n_elements(params))
;  for i=0,n_elements(params)-1,2 do begin
;    delta_params(*,i) = exp(-x/params[i+1])
;    delta_params(*,i) = exp(-x/params[i+1])
;  endfor
;endif

; Return resuls
;print,[params,model[0:3]]
return,model

end

;************************************************************

function acs_fit_cte_trail, y, y_error, $
                            trail_lengths=trail_lengths, $
                            n_species=n_species, $
                            trap_ratio=trap_ratio, $
                            positive_densities=positive_densities, $
                            additive_constant=additive_constant, $
                            silent=silent, $
                            plots=plots, yrange=yrange, noxlabels=noxlabels , noylabels=noylabels

;+
; NAME:
;      ACS_FIT_CTE_TRAIL
;
; CATEGORY:
;      CTE removal
;
; PURPOSE:
;      Fit one trail, behind one (mean of many) warm pixel(s). 
;
; INPUTS:
;      
;
; OPTIONAL INPUTS:
;      None.
;
; KEYWORD PARAMETERS:
;      None.
;
; OUTPUTS:
;      
;
; MODIFICATION HISTORY:
;      - Written by Richard Massey.
;- 

; Parse inputs
if not keyword_set(y_error) then y_error=replicate(0,n_elements(y))
if n_elements(y) ne n_elements(y_error) then message,"Inputs incorrectly formatted!"
x=indgen(n_elements(y))+1
if not keyword_set(n_species) then begin
  if keyword_set(trail_lengths) then n_species=n_elements(trail_lengths) else n_species=2
endif

; Set up initial guesses and constraints
parinfo=replicate({value:0.D, fixed:0B, limited:[0B,0B], limits:[0.D,0.D], step:1e-3, tied:""}, n_species*2+1)
parinfo[indgen(n_species)*2].value=1. ; Trap densities
if keyword_set(positive_densities) then parinfo[indgen(n_species)*2].limited=[1B,0B]
if keyword_set(trail_lengths) then begin ; Trap lifetimes
  parinfo[indgen(n_species)*2+1].value=trail_lengths
  parinfo[indgen(n_species)*2+1].fixed=1B
endif else begin
  parinfo[indgen(n_species)*2+1].value=[0.84,6.2]
  parinfo[indgen(n_species)*2+1].limited=[1B,1B]
  parinfo[indgen(n_species)*2+1].limits=[0.1,50]
  parinfo[indgen(n_species)*2+1].fixed=0B
endelse
if not keyword_set(additive_constant) then parinfo[n_species*2].fixed=1B else parinfo[n_species*2].limited=[0B,0B]
if keyword_set(trap_ratio) then for i=1,n_species-1 do parinfo[2*i].tied="P[0]*("+strtrim(trap_ratio[i]/float(trap_ratio[0]),2)+")"

; State initial guesses
if not keyword_set(silent) then print,[(["density","length"])[indgen(2*n_species) mod 2]+strtrim(1+indgen(2*n_species)/2,2),"constant"],FORMAT="(A36,8(A11))"
if not keyword_set(silent) then print,"Initial parameter guesses: ",parinfo.value,FORMAT="(A27,8(F9.6,:,', '))"

; Iterate to best parameters
params = mpfitfun("acs_cte_trail_model", float(x), float(y), float(y_error), PERROR=perror, PARINFO=parinfo, STATUS=status, ERRMSG=errmsg, FTOL=1e-7, YFIT=y_fit, /QUIET)
if keyword_set(trap_ratio) then for i=1,n_species-1 do perror[2*i]=perror[0]*trap_ratio[i]/float(trap_ratio[0])
if not keyword_set(silent) then print,"Best-fit parameters: ",params,FORMAT="(A27,8(F9.6,:,', '))"
if not keyword_set(silent) then print,"Errors on parameters: ",perror,FORMAT="(A27,8(F9.6,:,', '))"
if not keyword_set(silent) then print,"Fixed? ",parinfo.fixed,FORMAT="(A27,8(I9,:,', '))"
if not keyword_set(silent) then print," Tied? ",strlen(parinfo.tied)<1,FORMAT="(A27,8(I9,:,', '))"
if status ne 1 then begin
  message,"Iteration did not converge! (status="+strtrim(status,2)+") "+errmsg,/INFO
  ;y_fit=y-y
endif

; Draw plots
if keyword_set(plots) then begin
  if not keyword_set(yrange) then yrange=[0.9*min(abs(y)-y_error),1.1*max(abs(y)+y_error)]
;  if keyword_set(noxlabels) then xtickname=replicate(" ",60) else xtitle="!6Number of pixels !8i!6 away from warm pixel !8p!d0!6!n"
  if keyword_set(noxlabels) then xtickname=replicate(" ",60) else xtitle="!6Pixels !8i!6 away from warm pixel"
  if keyword_set(noylabels) then ytickname=replicate(" ",60) else ytitle="!6CTE trail !8T!Di!6!N [electrons]"
  plot,x,y_fit,psym=-3,linestyle=2,/ylog,/ystyle,yrange=yrange,xrange=[0.5,max(x)+0.5],/xstyle,$
       xtitle=xtitle,xtickname=xtickname,ytitle=ytitle,ytickname=ytickname;,xtickv=indgen(9)+1
  oplot,x,-y_fit,psym=-3,linestyle=1
  oplot,[min(x,max=maxx),maxx],replicate(params[n_species*2],2),linestyle=2,psym=-3
  usersym,[-0.5,0.5],[0,0] ; Set psym=8 to be a horizontal line
  for i=0,n_elements(x)-1 do oplot,replicate(x[i],2),yrange[0]>(abs(y[i])+[-1,1]*y_error[i])<yrange[1],psym=-8
  oplot,psym=-8,x,y
  usersym,cos(2*!pi*findgen(21)/20),sin(2*!pi*findgen(21)/20) ; Set psym=8 to be a circle
  oplot,psym=8,x,-y
endif

; Store results in one convenient structure
return,{x:x,y:y,y_error:y_error,y_fit:y_fit, $
        trail_lengths:params[indgen(n_species)*2+1], $
        trail_lengths_error:perror[indgen(n_species)*2+1], $
        trail_amplitudes:params[indgen(n_species)*2], $
        trail_amplitudes_error:perror[indgen(n_species)*2], $
        additive_constant:params[n_species*2], $
        additive_constant_error:perror[n_species*2], $
        status:status,errmsg:errmsg}

end

;************************************************************

pro acs_fit_cte_trails, trails, $
                        trail_lengths=trail_lengths, $
                        n_species=n_species, $
                        fixed_trap_ratio=fixed_trap_ratio, $
                        additive_constant=additive_constant, $
                        positive_densities=positive_densities, $
                        plots=plots, postscript=postscript
;+
; NAME:
;      ACS_FIT_CTE_TRAILS
;
; CATEGORY:
;      Shapelets.
;
; PURPOSE:
;      
;
; INPUTS:
;      
;
; OPTIONAL INPUTS:
;      None.
;
; KEYWORD PARAMETERS:
;      None.
;
; OUTPUTS:
;      
;
; MODIFICATION HISTORY:
;      - Written by Richard Massey.
;-

if tag_exist(trails,"TRAIL_AMPLITUDES") then goto, skippy

; Parse input and strip unwanted tags from trails structure
if not keyword_set(trails) then message,"TRAILS structure not specified!"
tagnames=tag_names(trails)
keep_tags=where(strmatch(tagnames,"TRAIL_LENGTHS",/FOLD_CASE)+strmatch(tagnames,"TRAIL_LENGTHS_ERROR",/FOLD_CASE)+$
                strmatch(tagnames,"TRAIL_AMPLITUDES",/FOLD_CASE)+strmatch(tagnames,"TRAIL_AMPLITUDES_ERROR",/FOLD_CASE)+$
                strmatch(tagnames,"ADDITIVE_CONSTANTS",/FOLD_CASE)+strmatch(tagnames,"ADDITIVE_CONSTANTS_ERROR",/FOLD_CASE)+$
                strmatch(tagnames,"METAPARAMETERS",/FOLD_CASE)+strmatch(tagnames,"METAPARAMETERS_ERROR",/FOLD_CASE) eq 0,n_keep_tags)
if n_keep_tags ne n_tags(trails) then begin
  new_trails=create_struct(tagnames[keep_tags[0]],trails.(keep_tags[0]))
  for i=1,n_keep_tags-1 do new_trails=create_struct(temporary(new_trails),tagnames[keep_tags[i]],trails.(keep_tags[i]))
  trails=temporary(new_trails)
endif
if keyword_set(fixed_trap_ratio) then if n_elements(fixed_trap_ratio) gt 1 then trap_ratio=fixed_trap_ratio
  

; Fit trail lengths
if not keyword_set(trail_lengths) then begin
  message,/INFO,"Fitting lengths of CTE trails using the brightest pixels furthest from the readout register."
  ti=trails.n_tran_bins-1
  fi=trails.n_flux_bins-2
  di=trails.n_date_bins-1
  bi=trails.n_back_bins-1
  if keyword_set(plots) and keyword_set(postscript) then ops,"acs_fit_cte_trail.eps",/encap,xsize=20,ysize=18
  xmargin=!x.margin & ymargin=!y.margin
  !x.margin=[6,0.25] & !y.margin=[3.25,0.25]
  trail=acs_fit_cte_trail(trail_lengths=[6.2,0.84],trap_ratio=trap_ratio,trails.trails[*,ti,fi,di,bi],trails.trails_error[*,ti,fi,di,bi],plots=plots,additive_constant=additive_constant,positive_densities=positive_densities)
  !x.margin=xmargin & !y.margin=ymargin
  if keyword_set(plots) and keyword_set(postscript) then cps
  trail_lengths=trail.trail_lengths
  trail_lengths_error=trail.trail_lengths_error
endif else trail_lengths_error=replicate(0,n_elements(trail_lengths))
trails=create_struct(temporary(trails),"TRAIL_LENGTHS",trail_lengths[sort(trail_lengths)],"TRAIL_LENGTHS_ERROR",trail_lengths_error[sort(trail_lengths)])
print,systime()
read,strtrim(trail_lengths,2)+": ",kk

; Fit trail amplitudes
n_species=n_elements(trails.trail_lengths)
for i=0,n_species-1 do message,/INFO,"Using trails of length "+strtrim(trails.trail_lengths[i],2)+"+/-"+strtrim(trails.trail_lengths_error[i],2)+" pixels."
trails=create_struct(trails,"TRAIL_AMPLITUDES",fltarr(n_species,trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins), $
                            "TRAIL_AMPLITUDES_ERROR",fltarr(n_species,trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins),$
                            "ADDITIVE_CONSTANTS",fltarr(n_species,trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins), $
                            "ADDITIVE_CONSTANTS_ERROR",fltarr(n_species,trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins))

; Start plot
if keyword_set(plots) then begin
  if keyword_set(postscript) then ops,"acs_fit_cte_trails.eps",/encap,xsize=60,ysize=35
  if strupcase(!d.name) eq "X" then window,0,title="CTE trails as a function of pixel flux and number of CCD transfers"
  !p.multi=[0,trails.n_flux_bins,trails.n_tran_bins,0,1]
  xmargin=!x.margin 
  ymargin=!y.margin
endif

; Fit trail amplitudes in each bin
for di=trails.n_date_bins-1,0,-1 do begin
  for bi=0,trails.n_back_bins-1 do begin
    ;trap_ratio=0
    for fi=trails.n_flux_bins-1,0,-1 do begin
      !p.multi[0]=5*(trails.n_flux_bins-fi)
      for ti=trails.n_tran_bins-1,0,-1 do begin
        !x.margin=9./(trails.n_flux_bins)*[trails.n_flux_bins-fi,-trails.n_flux_bins+fi+1]+[-0.25,0.25]
        !y.margin=4./(trails.n_tran_bins)*[trails.n_tran_bins-ti,-trails.n_tran_bins+ti+1]+[-0.25,0.25]
        if trails.n_trails[ti,fi,di,bi] gt 0 then begin
          trail=acs_fit_cte_trail(trails.trails[*,ti,fi,di,bi],trails.trails_error[*,ti,fi,di,bi],silent=0,plots=plots,yrange=[1e-2,90],$
                                  trail_lengths=trail_lengths,trap_ratio=trap_ratio,additive_constant=additive_constant,positive_densities=positive_densities,$
                                  noxlabels=(ti gt 0),noylabels=(fi gt 0))
          trails.trail_amplitudes[*,ti,fi,di,bi]=trail.trail_amplitudes
          trails.trail_amplitudes_error[*,ti,fi,di,bi]=trail.trail_amplitudes_error
          trails.additive_constants[*,ti,fi,di,bi]=trail.additive_constant
          trails.additive_constants_error[*,ti,fi,di,bi]=trail.additive_constant_error
          if keyword_set(fixed_trap_ratio) then begin
            if n_elements(fixed_trap_ratio) gt 1 then trap_ratio=fixed_trap_ratio else trap_ratio=trail.trail_amplitudes/float(trail.trail_amplitudes[0]) 
            trails.trail_amplitudes_error[*,ti,fi,di,bi]=trap_ratio*trails.trail_amplitudes_error[0,ti,fi,di,bi]
          endif
        endif
      endfor
    endfor
    if keyword_set(fixed_trap_ratio) then print,"Trap density ratio:",trap_ratio
  endfor
endfor

; Finish off plot
if keyword_set(plots) then begin
  !p.multi=0
  !x.margin=xmargin
  !y.margin=ymargin
  xyouts,0.995,0.59,"!6Hot pixels",charsize=2,/norm,align=1
  xyouts,0.03,0.59, "!6Warm pixels",charsize=2,/norm,align=0
  xyouts,0.51,0.04,"!6Near readout register",charsize=2,/norm,align=0.5
  xyouts,0.51,0.975,"!6Far from readout register",charsize=2,/norm,align=0.5
  if keyword_set(plots) and keyword_set(postscript) then cps
endif



print,systime()
read,kk
skippy:

; Fit CTE metamodel (First set up default guesses and limits)
n_species=n_elements(trails.trail_lengths)
parinfo = [{value:0.01D,fixed:0B, limited:[1B,0B], limits:[0.0D,0.0D], tied:""},$ ; Overall normalisation
           {value:0.0D, fixed:1B, limited:[1B,0B], limits:[0.0D,0.0D], tied:""},$ ; Rate at which the number of charge traps grows over time
           {value:1.0D, fixed:1B, limited:[0B,0B], limits:[0.0D,0.0D], tied:""},$ ; Power to which the date is raised
           {value:1.0D, fixed:1B, limited:[1B,1B], limits:[0.5D,1.5D], tied:""},$ ; Power to which the number of transfers is raised
           {value:2048D,fixed:1B, limited:[1B,0B], limits:[0.0D,0.0D], tied:""},$ ; Maximum number of transfers
           {value:200D, fixed:0B, limited:[1B,1B], limits:[0.0D,3000.0D], tied:""},$ ; Effective notch depth
           {value:double(trails.full_well_depth), fixed:1B, limited:[0B,0B], limits:[0.0D,0.0D], tied:""},$ ; Full well depth
           {value:0.5D, fixed:0B, limited:[1B,1B], limits:[0.1D,2.0D], tied:""},$ ; Power to which the flux is raised
           {value:0.5D, fixed:0B, limited:[0B,0B], limits:[0.0D,0.0D], tied:"P[7]"}]  ; Power to which the background level is raised
if trails.n_date_bins gt 1 then begin
  parinfo[1]={value:0.01D, fixed:0B, limited:[1B,0B], limits:[0.0D,0.0D], tied:""}
  parinfo[2]={value:1.0D,  fixed:1B, limited:[0B,0B], limits:[0.0D,0.0D], tied:""}
endif

; Fix the index alpha
parinfo[7]={value:0.46D,  fixed:1B, limited:[0B,0B], limits:[0.0D,0.0D], tied:""}
parinfo[8]={value:0.46D,  fixed:1B, limited:[0B,0B], limits:[0.0D,0.0D], tied:""}

; Fit each species separately (if they were tied earlier, this will produce identical results)
if not tag_exist(trails,"METAPARAMETERS") then trails=create_struct(trails,"METAPARAMETERS",fltarr(n_species,n_elements(parinfo)),"METAPARAMETERS_ERROR",fltarr(n_species,n_elements(parinfo)))
for ii=1-keyword_set(fixed_trap_ratio),0,-1 do begin ; Do twice so we can make sure the notch depth is the same for all species of charge trap
  for i=0,n_species-1 do begin
    message,/INFO,"Fitting metaparameters for the species of charge trap with decay time="+strtrim(trails.trail_lengths[i],2)+" pixels."

    ; Fit well filling rate
    print,(["norm","growth","date_pow","tran_pow","max_tran","notch_depth","full_well","flux_pow","back_pow"])[0:n_elements(parinfo)-1],FORMAT="(A37,9(A12))"
    print,"Initial parameter guesses: ",parinfo.value,FORMAT="(A27,9(F10.5,:,', '))"
    x=[[reform(trails.tran,n_elements(trails.tran))],[reform(trails.flux,n_elements(trails.flux))],[reform(trails.date,n_elements(trails.date))],[reform(trails.back,n_elements(trails.back))]]
    y=reform(trails.trail_amplitudes[i,*,*,*,*],n_elements(trails.trail_amplitudes[i,*,*,*,*]))
    y_error=reform(trails.trail_amplitudes_error[i,*,*,*,*],n_elements(trails.trail_amplitudes_error[i,*,*,*,*]))
    downweight_high_flux=100. ; Fudge to downweight high flux bin, since there is actually a lot of error in x
    y_error[n_elements(y)-trails.n_tran_bins:n_elements(y)-1]=y_error[n_elements(y)-trails.n_tran_bins:n_elements(y)-1]*downweight_high_flux
    params=mpfitfun("acs_cte_model",x,y,y_error,PERROR=perror,PARINFO=parinfo,STATUS=status,ERRMSG=errmsg,YFIT=y_fit,/QUIET)
    y_error[n_elements(y)-trails.n_tran_bins:n_elements(y)-1]=y_error[n_elements(y)-trails.n_tran_bins:n_elements(y)-1]/downweight_high_flux
    print,"Best-fit parameters: ",params,FORMAT="(A27,9(F10.5,:,', '))"
    print,"Errors on parameters: ",perror,FORMAT="(A27,9(F10.5,:,', '))"
    print,"Fixed? ",parinfo.fixed,FORMAT="(A27,9(I9,:,', '))"
    print," Tied? ",strlen(parinfo.tied)<1,FORMAT="(A27,9(I9,:,', '))"
    if status ne 1 then message,"Iteration did not converge! (status="+strtrim(status,2)+") "+errmsg,/INFO
    trails.metaparameters[i,*]=params & trails.metaparameters_error[i,*]=perror
    chi_squared=mean(((y_fit-y)/y_error)^2) & print,chi_squared
    bf_density=params[0]
    bf_density_growth=params[1]
    bf_density_power=params[2]
    bf_tran_power=params[3]
    bf_n_pixels=params[4]
    bf_notch_depth=params[5]
    bf_full_well_depth=params[6]
    bf_flux_power=params[7]
    bf_back_power=params[8]

    ; Start plot
    if ii eq 0 and keyword_set(plots) then begin
      ; Adjust variables so they are more convenient to plot
      x=reform(x[*,1],trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins)
      y=reform(y,trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins)
      y_error=reform(y_error,trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins)
      y_fit=reform(y_fit,trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins)
      xrange=[0.9*bf_notch_depth<min(x),max(x)*1.02]
      yrange=[acs_cte_model(transpose([1,bf_notch_depth,mean(trails.date),mean(trails.back)]),params),max(abs(y)+y_error)*1.1] & yrange[0]=yrange[0]-yrange[1]/10
      ; Make plot
      if keyword_set(postscript) then ops,"acs_well_filling"+strtrim(trails.trail_lengths[i],2)+".eps",/encap,xsize=30,ysize=35
      if strupcase(!d.name) eq "X" then begin & window,i,title="Well filling rate" & !p.multi=[0,2,1,0,0] & endif
      xmargin=!x.margin 
      ymargin=!y.margin
      for di=0,trails.n_date_bins-1 do begin
        for bi=0,trails.n_back_bins-1 do begin
          plot,replicate(bf_notch_depth,2),[-1,1]*1e9,psym=-3,xrange=xrange,yrange=yrange,/xstyle,/ystyle,linestyle=1,$
               ;title="!6Charge traps with !7s!6="+strtrim(trails.trail_lengths[i],2)+" pixels",$
               xtitle="!6Number of electrons !8n!6 in a pixel",ytitle="!6Number of exposed traps between a pixel and the readout register !8q!6"
          xyouts,orientation=90,bf_notch_depth-(xrange[1]-xrange[0])/200.,yrange[1]*0.95,align=1,"!6(Fitted) notch depth !8d!6"
          oplot,[-bf_full_well_depth,bf_notch_depth],[0,0],psym=-3,linestyle=1 ; Theoretical model
          for ti=0,trails.n_tran_bins-1 do begin
            n_xx=100
            xx=bf_notch_depth+findgen(n_xx)/(n_xx-1)*(xrange[1]-xrange[0])
            yy=acs_cte_model([[replicate(mean(trails.tran[ti,*,di,bi]),n_xx)],[xx],[replicate(mean(trails.date[ti,*,di,bi]),n_xx)],[replicate(mean(trails.back[ti,*,di,bi]),n_xx)]],params)
            oplot,xx,yy,psym=-3,linestyle=1
            for fi=0,trails.n_flux_bins-1 do begin
              usersym,[-0.5,0.5],[0,0] ; Set psym=8 to be a horizontal line
              oplot,replicate(trails.flux[ti,fi,di,bi],2),y[ti,fi,di,bi]+[-1,1]*y_error[ti,fi,di,bi],psym=-8 ; Vertical error bars
              usersym,[0,0],[-0.5,0.5] ; Set psym=8 to be a vertical line
              if fi ne trails.n_flux_bins-1 then oplot,trails.flux[ti,fi,di,bi]+[-1,1]*trails.flux_rms[ti,fi,di,bi],replicate(y[ti,fi,di,bi],2),psym=-8 ; Horizontal error bars
            endfor
            oplot,x[ti,*,di,bi],y[ti,*,di,bi],psym=-3 ; Line joining data points
          endfor
        endfor
      endfor
      if keyword_set(postscript) then cps
    endif
  endfor
  ; Fix notch depth
  parinfo[5].value=mean(trails.metaparameters[*,5]) & parinfo[5].fixed=1B
  parinfo[5].value=max(trails.metaparameters[*,5]) & parinfo[5].fixed=1B
endfor

; Finish off plot
if keyword_set(plots) then begin
  !p.multi=0
  !x.margin=xmargin
  !y.margin=ymargin
endif

;; Add results to structure
;tags=tag_names(trails)
;duplicate_tags=strmatch(tags,"METAPARAMETERS")+strmatch(tags,"METAPARAMETERS_ERROR")+strmatch(tags,"TRAIL_AMPLITUDES_MODEL")
;if total(duplicate_tags) gt 0 then begin ; Remove duplicate tags
;  str=""
;  good_tags=where(duplicate_tags eq 0,n_good_tags)
;  for i=0,n_good_tags-1 do str=str+","+tags[good_tags[i]]+":trails.("+strtrim(good_tags[i],2)+")"
;  junk=execute("trails={"+strmid(str,1)+"}")
;endif
;trails=create_struct(temporary(trails),"METAPARAMETERS",metaparameters,"METAPARAMETERS_ERROR",metaparameters_error,"TRAIL_AMPLITUDES_MODEL",trail_amplitudes_model)


end
