function acs_cte_model, x, params

; NAME:
;      ACS_CTE_MODEL
;
; CATEGORY:
;      Pixel-by-pixel correction of trailing due to charge transfer inefficiency in CCD detectors.
;
; PURPOSE:
;      A model for a CTE model as a function of date, pixel flux, CCD position etc.
;
; INPUTS:
;      X      - An Nx4 floating point array, where the 4 are [N_TRANS,FLUX,DATE,BACKGROUND]
;      PARAMS - A 3, 4 or 5-element floating point array (one more than the number of variables in X).
;               The first parameter is an overall normalisation (i.e. proportional to mean trap density)
;               The second parameter is the power to which the number of transfers is raised
;               The third parameter is the power to which the flux is raised
;               The fourth parameter is the effective notch depth
;               The fifth parameter is the growth rate of traps per day; it is not used (and should not be passed) if MJD is not supplied 
;               The sixth parameter is the power to which the background level is raised; it is not used (and should not be passed) if background level is not supplied 
;
; OPTIONAL INPUTS:
;      None.
;
; KEYWORD PARAMETERS:
;      
;
; OUTPUTS:
;      
;
; MODIFICATION HISTORY:
;      Jan 08 - Written by Richard Massey.

; Extract variables into nicer local names
density=params[0]
growth_rate=params[1]
date_power=params[2]
tran_power=params[3]
n_pixels=params[4]
notch_depth=params[5] & if notch_depth le 1 then notch_depth=0
full_well_depth=params[6]
flux_power=params[7]
back_power=params[8]

; Extract ordinates
tran=x[*,0]
flux=x[*,1]
date=x[*,2]
back=x[*,3]

; Compute model
return, (density+growth_rate*date^date_power) * (tran^tran_power) $
       *(   (((flux-notch_depth)>0)/float(full_well_depth-notch_depth))^flux_power $
           -(((back-notch_depth)>0)/float(full_well_depth-notch_depth))^back_power  )>0
end

;************************************************************

function acs_cte_trail_model, x, params, delta_params

; NAME:
;      ACS_FIT_TRAIL_MODEL
;
; PURPOSE:
;      A multiple exponential model for a CTE trail.
;
; INPUTS:
;      X      -
;      PARAMS - floating point vector with an even number of elements that
;               alternately contain an amplitude and an e-folding factor of
;               one exponential.
;
; MODIFICATION HISTORY:
;      Jan 08 - Written by Richard Massey.

; Check input parameters are acceptable
if n_elements(params) eq 0 then message,"Must specify some model parameters!"
;if min(params) lt 0 then print,"WARNING: negative parameter ",params

;print,x
;i=0 & print,params[i+1]
;print,total(exp((-indgen(10*params[i+1]))/float(params[i+1]))),exp(1/float(params[i+1]))-1
;print,exp(-x/float(params[i+1])) / total(exp((-0.5-indgen(10*params[i+1]))/float(params[i+1])))
;print,exp(-x/float(params[i+1])) * (exp(1/float(params[i+1]))-1)
;print,(exp(-x/float(params[i+1])) / total(exp((-0.5-indgen(10*params[i+1]))/float(params[i+1]))))/(exp(-x/float(params[i+1])) * (exp(1/float(params[i+1]))-1))
;i=2 & print,params[i+1]
;print,total(exp((-indgen(10*params[i+1]))/float(params[i+1]))),exp(1/float(params[i+1]))-1
;print,exp(-x/float(params[i+1])) / total(exp((-0.5-indgen(10*params[i+1]))/float(params[i+1])))
;print,exp(-x/float(params[i+1])) * (exp(1/float(params[i+1]))-1)
;print,(exp(-x/float(params[i+1])) / total(exp((-0.5-indgen(10*params[i+1]))/float(params[i+1]))))/(exp(-x/float(params[i+1])) * (exp(1/float(params[i+1]))-1))


; Compute model using these parameters
;model=replicate(x[0]*0.,n_elements(x))
;for i=0,n_elements(params)-(n_elements(params) mod 2)-1,2 do $
;  model=model+params[i]*exp(-x/float(params[i+1])) $
;                  / total(exp((-0.5-indgen(10*params[i+1]))/float(params[i+1])))
model=x*0.;replicate(x[0]*0.,n_elements(x))
for i=0,n_elements(params)-(n_elements(params) mod 2)-1,2 do $
  model=model+params[i]*exp(-x/float(params[i+1])) $
                  / total(exp((-1-indgen(100*params[i+1]))/float(params[i+1])))
if (n_elements(params) mod 2) eq 1 then model=model+(reverse(params))[0]

;; Compute analytic derivatives
;if n_params() GT 2 then begin
;  delta_params = replicate(x[0],n_elements(x),n_elements(params))
;  for i=0,n_elements(params)-1,2 do begin
;    delta_params(*,i) = exp(-x/params[i+1])
;    delta_params(*,i) = exp(-x/params[i+1])
;  endfor
;endif

; Return resuls
;print,[params,model[0:3]]
return,model

end

;************************************************************

function acs_fit_cte_trail, y, y_error, $
                            trail_lengths=trail_lengths, $
                            n_species=n_species, $
                            trap_ratio=trap_ratio, $
                            positive_densities=positive_densities, $
                            additive_constant=additive_constant, $
                            silent=silent, $
                            plots=plots, yrange=yrange, noxlabels=noxlabels , noylabels=noylabels,color=color

;+
; NAME:
;      ACS_FIT_CTE_TRAIL
;
; CATEGORY:
;      CTE removal
;
; PURPOSE:
;      Fit one trail, behind one (mean of many) warm pixel(s). 
;
; INPUTS:
;      
;
; OPTIONAL INPUTS:
;      None.
;
; KEYWORD PARAMETERS:
;      None.
;
; OUTPUTS:
;      
;
; MODIFICATION HISTORY:
;      - Written by Richard Massey.
;- 

; Parse inputs
if not keyword_set(y_error) then y_error=replicate(0,n_elements(y))
if n_elements(y) ne n_elements(y_error) then message,"Inputs incorrectly formatted!"
x=indgen(n_elements(y))+1
if not keyword_set(n_species) then begin
  if keyword_set(trail_lengths) then n_species=n_elements(trail_lengths) else n_species=2
endif

; Set up initial guesses and constraints
parinfo=replicate({value:0.D, fixed:0B, limited:[0B,0B], limits:[0.D,0.D], step:1e-3, tied:""}, n_species*2+1)
parinfo[indgen(n_species)*2].value=1. ; Trap densities
if keyword_set(positive_densities) then parinfo[indgen(n_species)*2].limited=[1B,0B]
if keyword_set(trail_lengths) then begin ; Trap lifetimes
  parinfo[indgen(n_species)*2+1].value=trail_lengths
  parinfo[indgen(n_species)*2+1].fixed=1B
endif else begin
  parinfo[indgen(n_species)*2+1].value=[0.5,5.0]
  parinfo[indgen(n_species)*2+1].limited=[1B,1B]
  parinfo[indgen(n_species)*2+1].limits=[0.1,50]
  parinfo[indgen(n_species)*2+1].fixed=0B
endelse
if not keyword_set(additive_constant) then parinfo[n_species*2].fixed=1B else parinfo[n_species*2].limited=[0B,0B]
if keyword_set(trap_ratio) then for i=1,n_species-1 do parinfo[2*i].tied="P[0]*("+strtrim(trap_ratio[i]/float(trap_ratio[0]),2)+")"

; State initial guesses
if not keyword_set(silent) then begin
  print,[(["density","length"])[indgen(2*n_species) mod 2]+strtrim(1+indgen(2*n_species)/2,2),"constant"],FORMAT="(A36,8(A11))"
  print,"Initial parameter guesses: ",parinfo.value,FORMAT="(A27,8(F9.6,:,', '))"
endif

; Iterate to best parameters
params = mpfitfun("acs_cte_trail_model", float(x), float(y), float(y_error), PERROR=perror, PARINFO=parinfo, STATUS=status, ERRMSG=errmsg, FTOL=1e-7, YFIT=y_fit, /QUIET)
;if status lt 0 then stop;perror=fltarr(n_species*2+1)
if keyword_set(trap_ratio) then for i=1,n_species-1 do perror[2*i]=perror[0]*trap_ratio[i]/float(trap_ratio[0])
if not keyword_set(silent) then begin
  print,"Best-fit parameters: ",params,FORMAT="(A27,8(F9.4,:,', '))"
  print,"Errors on parameters: ",perror,FORMAT="(A27,8(F9.4,:,', '))"
  print,"Fixed? ",(["-","Y"])[parinfo.fixed],FORMAT="(A27,8(A9,:,', '))"
  print," Tied? ",(["-","Y"])[strlen(parinfo.tied)<1],FORMAT="(A27,8(A9,:,', '))"
  print,status,errmsg
endif
if status ne 1 then begin
  message,"Iteration did not converge! (status="+strtrim(status,2)+") "+errmsg,/INFO
  ;y_fit=y-y
endif

; Draw plots
if keyword_set(plots) then begin
  if not keyword_set(yrange) then yrange=[0.9*min(abs(abs(y)-y_error)),1.1*max(abs(y)+y_error)]
;  if keyword_set(noxlabels) then xtickname=replicate(" ",60) else xtitle="!6Number of pixels !8i!6 away from warm pixel !8p!d0!6!n"
  if keyword_set(noxlabels) then xtickname=replicate(" ",60) else xtitle="!6Pixels !8i!6 away from warm pixel"
  if keyword_set(noylabels) then ytickname=replicate(" ",60) else ytitle="!6CTE trail !8T!Di!6!N [electrons]"
  plot,[0,0],/NODATA,psym=-3,linestyle=2,/ylog,/ystyle,yrange=yrange,xrange=[0.5,max(x)+0.5],/xstyle,$
       xtitle=xtitle,xtickname=xtickname,ytitle=ytitle,ytickname=ytickname;,color=color;,xtickv=indgen(9)+1
  oplot,x,y_fit,psym=-3,linestyle=2,color=color
  oplot,x,-y_fit,psym=-3,linestyle=1,color=color
  oplot,[min(x,max=maxx),maxx],replicate(params[n_species*2],2),linestyle=2,psym=-3,color=color
  usersym,[-0.5,0.5],[0,0] ; Set psym=8 to be a horizontal line
  for i=0,n_elements(x)-1 do oplot,replicate(x[i],2),yrange[0]>(abs(y[i])+[-1,1]*y_error[i])<yrange[1],psym=-8,color=color
  oplot,psym=-8,x,y,color=color
  usersym,cos(2*!pi*findgen(21)/20),sin(2*!pi*findgen(21)/20) ; Set psym=8 to be a circle
  oplot,psym=8,x,-y,color=color
endif

; Store results in one convenient structure
return,{x:x,y:y,y_error:y_error,y_fit:y_fit, $
        trail_lengths:params[indgen(n_species)*2+1], $
        trail_lengths_error:perror[indgen(n_species)*2+1], $
        trail_amplitudes:params[indgen(n_species)*2], $
        trail_amplitudes_error:perror[indgen(n_species)*2], $
        additive_constant:params[n_species*2], $
        additive_constant_error:perror[n_species*2], $
        status:status,errmsg:errmsg}

end

;************************************************************

pro acs_fit_cte_trails, trails, $
                        trap_ratio=trap_ratio,$
                        additive_constant=additive_constant,$
                        positive_densities=positive_densities,$
                        plots=plots, postscript=postscript
                        
; NAME:
;      ACS_FIT_CTE_TRAILS
;
; CATEGORY:
;      Shapelets.
;
; PURPOSE:
;      
;
; INPUTS:
;      
;
; OPTIONAL INPUTS:
;      None.
;
; KEYWORD PARAMETERS:
;      None.
;
; OUTPUTS:
;      
;
; MODIFICATION HISTORY:
;      - Written by Richard Massey.


; Parse input and strip unwanted tags from trails structure
if not keyword_set(trails) then message,"TRAILS structure not specified!"

; Start plot
if keyword_set(plots) then begin
  xmargin=!x.margin 
  ymargin=!y.margin
  if keyword_set(postscript) then ops,"acs_cte_trails_"+postscript+".eps",/encap,xsize=60,ysize=35
  if strupcase(!d.name) eq "X" then window,0,title="CTE trails as a function of pixel flux and number of CCD transfers"
  !p.multi=[0,trails.n_flux_bins,trails.n_tran_bins,0,1]
endif

; Fit trail amplitudes in each bin
for di=trails.n_date_bins-1,0,-1 do begin
  for bi=0,trails.n_back_bins-1 do begin
    ;trap_ratio=0
    for fi=trails.n_flux_bins-1,0,-1 do begin
      !p.multi[0]=trails.n_tran_bins*(trails.n_flux_bins-fi)
      for ti=trails.n_tran_bins-1,0,-1 do begin
        !x.margin=9./(trails.n_flux_bins)*[trails.n_flux_bins-fi,-trails.n_flux_bins+fi+1]+[-0.25,0.25]
        !y.margin=4./(trails.n_tran_bins)*[trails.n_tran_bins-ti,-trails.n_tran_bins+ti+1]+[-0.25,0.25]
        if trails.n_trails[ti,fi,di,bi] gt 2 then begin
          trail=acs_fit_cte_trail(trails.trails[*,ti,fi,di,bi],trails.trails_error[*,ti,fi,di,bi],silent=1,plots=plots,yrange=[1e-2,300],$
                                  trail_lengths=trails.trail_lengths,trap_ratio=trap_ratio,additive_constant=additive_constant,positive_densities=positive_densities,$
                                  noxlabels=(ti gt 0),noylabels=(fi gt 0),color=(255+(bi-trails.n_back_bins+1)*240./trails.n_back_bins)*(1-keyword_set(postscript)))
          trails.trail_amplitudes[*,ti,fi,di,bi]=trail.trail_amplitudes
          trails.trail_amplitudes_error[*,ti,fi,di,bi]=trail.trail_amplitudes_error
          trails.additive_constants[*,ti,fi,di,bi]=trail.additive_constant
          trails.additive_constants_error[*,ti,fi,di,bi]=trail.additive_constant_error
         ;print,trail.trail_amplitudes[*]/trail.trail_amplitudes[0]
         ;if keyword_set(trap_ratio) then trails.trail_amplitudes_error[*,ti,fi,di,bi]=trap_ratio*trails.trail_amplitudes_error[0,ti,fi,di,bi]
        endif else begin
          trails.trail_amplitudes_error[*,ti,fi,di,bi]=!values.f_infinity
          trails.additive_constants_error[*,ti,fi,di,bi]=!values.f_infinity
          !p.multi[0]=!p.multi[0]-1
        endelse
      endfor
    endfor
    if keyword_set(trap_ratio) then print,"Trap density ratio:",trap_ratio
  endfor
endfor

; Finish off plot
!p.multi=0
if keyword_set(plots) then begin
  xyouts,0.995,0.59,"!6Hot pixels",charsize=2,/norm,align=1
  xyouts,0.03,0.59, "!6Warm pixels",charsize=2,/norm,align=0
  xyouts,0.51,0.04, "!6Near readout register",charsize=2,/norm,align=0.5
  xyouts,0.51,0.975,"!6Far from readout register",charsize=2,/norm,align=0.5
  if keyword_set(plots) and keyword_set(postscript) then cps
  !x.margin=xmargin
  !y.margin=ymargin
endif

end

;************************************************************

pro acs_fit_cte_model, trails, $
                       fixed_trap_ratio=fixed_trap_ratio, $
                       notch_depth=notch_depth, $
                       well_fill_power=well_fill_power, $
                       noplottheory=noplottheory, $
                       linear_algebra=linear_algebra, $
                       plots=plots, postscript=postscript
                        
; NAME:
;      ACS_FIT_CTE_MODEL
;
; CATEGORY:
;      Shapelets.
;
; PURPOSE:
;      
;
; INPUTS:
;      
;
; OPTIONAL INPUTS:
;      None.
;
; KEYWORD PARAMETERS:
;      None.
;
; OUTPUTS:
;      
;
; MODIFICATION HISTORY:
;      - Written by Richard Massey.

; Set up initial guesses
n_species=n_elements(trails.trail_lengths)
; Starting guess changed 15/1/09
;parinfo = [{value:0.1D, fixed:0B, limited:[1B,0B], limits:[0.0D,0.0D], tied:""},$ ; Overall normalisation
parinfo = [{value:0.1D, fixed:0B, limited:[1B,0B], limits:[0.0D,0.0D], tied:""},$ ; Overall normalisation
           {value:0.0D, fixed:1B, limited:[1B,0B], limits:[0.0D,0.0D], tied:""},$ ; Rate at which the number of charge traps grows over time
           {value:1.0D, fixed:1B, limited:[0B,0B], limits:[0.0D,0.0D], tied:""},$ ; Power to which the date is raised
           {value:1.0D, fixed:1B, limited:[1B,1B], limits:[0.5D,1.5D], tied:""},$ ; Power to which the number of transfers is raised
           {value:2048D,fixed:1B, limited:[1B,0B], limits:[0.0D,0.0D], tied:""},$ ; Maximum number of transfers
           {value:100D, fixed:0B, limited:[1B,1B], limits:[0.0D,1000.0D], tied:""},$ ; Effective notch depth
           {value:double(trails.full_well_depth), fixed:1B, limited:[0B,0B], limits:[0.0D,0.0D], tied:""},$ ; Full well depth
           {value:0.5D, fixed:0B, limited:[1B,1B], limits:[0.1D,2.0D], tied:""},$ ; Power to which the flux is raised
           {value:0.5D, fixed:0B, limited:[0B,0B], limits:[0.0D,0.0D], tied:"P[7]"}]  ; Power to which the background level is raised
if trails.n_date_bins gt 1 then begin
  parinfo[1]={value:0.01D, fixed:0B, limited:[1B,0B], limits:[0.0D,0.0D], tied:""}
  parinfo[2]={value:1.0D,  fixed:1B, limited:[0B,0B], limits:[0.0D,0.0D], tied:""}
endif
if keyword_set(notch_depth)     then parinfo[5]={value:double(notch_depth),     fixed:1B, limited:[0B,0B], limits:[0.0D,0.0D], tied:""}
if keyword_set(well_fill_power) then parinfo[7]={value:double(well_fill_power), fixed:1B, limited:[0B,0B], limits:[0.0D,0.0D], tied:""}


; Fit each species separately (if they were tied earlier, this should produce identical results... although due to numerical instabilities, it's sometimes a bit off)
if not tag_exist(trails,"METAPARAMETERS") then trails=create_struct(trails,"METAPARAMETERS",fltarr(n_species,n_elements(parinfo)),"METAPARAMETERS_ERROR",fltarr(n_species,n_elements(parinfo)))
for ii=1-keyword_set(fixed_trap_ratio),0,-1 do begin ; Do twice so we can make sure the notch depth is the same for all species of charge trap
  ;for i=0,n_species-1 do begin
  for i=0,n_species do begin
    if i lt n_species then begin
      message,/INFO,"Fitting metaparameters for the species of charge trap with decay time="+strtrim(trails.trail_lengths[i],2)+" pixels."
    endif else begin
      message,/INFO,"Total density of charge traps: "+strtrim(total(trails.metaparameters[*,0]),2)+"+/-"+strtrim(total(trails.metaparameters_error[*,0]),2)+" per pixel"
      message,/INFO,"Refitting total charge trap density"
    endelse

    ; Fit well filling rate
    print,(["norm","growth","date_pow","tran_pow","max_tran","notch_depth","full_well","flux_pow","back_pow"])[0:n_elements(parinfo)-1],FORMAT="(A37,9(A12))"
    print,"Initial parameter guesses: ",parinfo.value,FORMAT="(A27,9(F10.4,:,', '))"
    ; Interate in order to compensate for the flux missing from warm pixels when they are measured, because it had been moved in to the trail
    n_iterations=4
    for j=0,n_iterations do begin
      x=[[reform(trails.tran,n_elements(trails.tran))],[reform(trails.flux,n_elements(trails.flux))],[reform(trails.date,n_elements(trails.date))],[reform(trails.back,n_elements(trails.back))]]
      if j gt 0 then begin
        corrected_flux=trails.flux+acs_cte_model(x,params)
        corrected_back=trails.back;+acs_cte_model([[reform(trails.tran,n_elements(trails.tran))],[reform(trails.back,n_elements(trails.back))],[reform(trails.date,n_elements(trails.date))],[reform(trails.back,n_elements(trails.back))]],params)
        x=[[reform(trails.tran,n_elements(trails.tran))],[reform(corrected_flux,n_elements(corrected_flux))],[reform(trails.date,n_elements(trails.date))],[reform(corrected_back,n_elements(corrected_back))]]
      endif
      if i lt n_species then begin
        y=reform(trails.trail_amplitudes[i,*,*,*,*],n_elements(trails.trail_amplitudes[i,*,*,*,*]))
        y_error=reform(trails.trail_amplitudes_error[i,*,*,*,*],n_elements(trails.trail_amplitudes_error[i,*,*,*,*]))      
      endif else begin
        y=reform(total(trails.trail_amplitudes,1));,n_elements(trails.trail_amplitudes[0,*,*,*,*]))
        y_error=reform(total(trails.trail_amplitudes_error,1));,n_elements(trails.trail_amplitudes_error[0,*,*,*,*]))      
      endelse
      zero_error=where(abs(y_error) lt 1e-7,n_zero_error) & if n_zero_error gt 0 then y_error[zero_error]=!values.f_infinity ; Error catching
      downweight_low_flux=1;00. ; Fudge to downweight low flux bin, since there is contamination from peaks in read noise
      downweight_high_flux=1;00. ; Fudge to downweight high flux bin, since there is actually a lot of error in x
      y_error[0:trails.n_tran_bins-1]=y_error[0:trails.n_tran_bins-1]*downweight_low_flux
      y_error[n_elements(y)-trails.n_tran_bins:n_elements(y)-1]=y_error[n_elements(y)-trails.n_tran_bins:n_elements(y)-1]*downweight_high_flux
      params=mpfitfun("acs_cte_model",x,y,y_error,PERROR=perror,PARINFO=parinfo,STATUS=status,ERRMSG=errmsg,YFIT=y_fit,/QUIET,/NOCATCH)
      y_error[0:trails.n_tran_bins-1]=y_error[0:trails.n_tran_bins-1]/downweight_low_flux
      y_error[n_elements(y)-trails.n_tran_bins:n_elements(y)-1]=y_error[n_elements(y)-trails.n_tran_bins:n_elements(y)-1]/downweight_high_flux
      print,"Best-fit parameters: ",params,FORMAT="(A27,9(F10.4,:,', '))"
    endfor

    ; Redo fit with simple linear algebra
    if keyword_set(linear_algebra) and total((parinfo.fixed>strlen(parinfo.tied))<1) eq (n_elements(parinfo)-1) then begin
      print,"Errors on parameters: ",perror,FORMAT="(A27,9(F10.4,:,', '))"
      print,"Redoing fit with simple linear algebra"
      print,"Initial parameter guesses: ",params,FORMAT="(A27,9(F10.4,:,', '))"
      ;params[0]=1.
      for j=0,n_iterations do begin
        params[0]=1.
        x=[[reform(trails.tran,n_elements(trails.tran))],[reform(trails.flux,n_elements(trails.flux))],[reform(trails.date,n_elements(trails.date))],[reform(trails.back,n_elements(trails.back))]]
        if j gt 0 then begin
          corrected_flux=trails.flux+acs_cte_model(x,params)
          corrected_back=trails.back;+acs_cte_model([[reform(trails.tran,n_elements(trails.tran))],[reform(trails.back,n_elements(trails.back))],[reform(trails.date,n_elements(trails.date))],[reform(trails.back,n_elements(trails.back))]],params)
          x=[[reform(trails.tran,n_elements(trails.tran))],[reform(corrected_flux,n_elements(corrected_flux))],[reform(trails.date,n_elements(trails.date))],[reform(corrected_back,n_elements(corrected_back))]]
        endif
        if i lt n_species then begin
          y=reform(trails.trail_amplitudes[i,*,*,*,*])
          y_error=reform(trails.trail_amplitudes_error[i,*,*,*,*])
        endif else begin
          y=reform(total(trails.trail_amplitudes,1))
          y_error=reform(total(trails.trail_amplitudes_error,1))
        endelse
        good=where(finite(y_error))
        y_model=acs_cte_model(x, params)
        y_weight=(y_model/reform(y_error))^2
        params[0]=params[0]*total(y_weight[good]*y[good]/y_model[good])/total(y_weight[good])
        perror[0]=sqrt(total((reform(y_error[good])/y_model[good])^2))/sqrt(n_elements(y[good]))
        y_fit=y_model*params[0]
        print,"Best-fit parameters: ",params,FORMAT="(A27,9(F10.4,:,', '))"
      endfor
    endif
    
    if i lt n_species then begin
      trails.metaparameters[i,*]=params
      trails.metaparameters_error[i,*]=perror
    endif
    print,"Errors on parameters: ",perror,FORMAT="(A27,9(F10.4,:,', '))"
    print,"Fixed? ",(["-","Y"])[parinfo.fixed],FORMAT="(A27,9(A10,:,', '))"
    print," Tied? ",(["-","Y"])[strlen(parinfo.tied)<1],FORMAT="(A27,9(A10,:,', '))"
    if status ne 1 then message,"Iteration did not converge! (status="+strtrim(status,2)+") "+errmsg,/INFO
    message,/INFO,"Reduced chi squared = "+strtrim(total(((y_fit-y)/y_error)^2)/(n_elements(y)-n_elements(where(parinfo.fixed+strlen(parinfo.tied)<1 eq 0))),2)
    ; Manually check chi squared is at a minimum
    ;a=findgen(301)/100 & b=fltarr(301) & for iii=0,300 do b[iii]=total(((y_fit*a[iii]-y)/y_error)^2)/(n_elements(y)-n_elements(where(parinfo.fixed+strlen(parinfo.tied)<1 eq 0)))
    ;plot,a,b,psym=-3,/ylog

    
    ; Extract answers into user-friendly variables
    bf_density=params[0]
    bf_density_growth=params[1]
    bf_density_power=params[2]
    bf_tran_power=params[3]
    bf_n_pixels=params[4]
    bf_notch_depth=params[5]
    bf_full_well_depth=params[6]
    bf_flux_power=params[7]
    bf_back_power=params[8]

    ; Fix notch depth
    if keyword_set(fixed_trap_ratio) and i eq 0 then begin
      print,"Fixing notch depth at "+strtrim(trails.metaparameters[0,5],2)+" and well fill power at "+strtrim(trails.metaparameters[0,7],2)
      parinfo[5].value=trails.metaparameters[0,5] & parinfo[5].fixed=1B
      parinfo[7].value=trails.metaparameters[0,7] & parinfo[7].fixed=1B
    endif

    if ii eq 0 and keyword_set(plots) then begin
      ; Start plot
      ; Adjust variables so they are more convenient to plot
      x=reform(x[*,1],trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins)
      y=reform(y,trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins);*4
      y_error=reform(y_error,trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins);*4
      y_fit=reform(y_fit,trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins);*4
      if bf_notch_depth le 1 then xrange=[min(x)/1.25,max(x)*1.25] else xrange=[0.9*bf_notch_depth<min(x),max(x)*1.25]
      yrange=[acs_cte_model(transpose([1,bf_notch_depth,mean(trails.date),mean(trails.back)]),params),max((abs(y)+y_error)[where(y_error ne !values.f_infinity)])*1.1] & yrange[0]=yrange[0]-yrange[1]/10
      ;xrange=[2630,2650] & yrange=[0.01,10]
      ;xrange=[100,30000] & 
      yrange=[0.3,300]
      ; Make plot
      if keyword_set(postscript) then ops,"acs_well_filling_"+postscript+"_"+([strtrim(trails.trail_lengths,2),"all"])[i]+".eps",/encap,xsize=30,ysize=35
      if strupcase(!d.name) eq "X" then begin & window,i,title="Well filling rate" & !p.multi=[0,2,1,0,0] & endif
      xmargin=!x.margin & !x.margin=[7,0.5]
      ymargin=!y.margin & !y.margin=[3.2,0.5]
      plot,[0,0],/nodata,xrange=xrange,yrange=yrange,/xstyle,/ystyle,/xlog,ylog=1,$
        ;title="!6Charge traps with !7s!6="+strtrim(trails.trail_lengths[i],2)+" pixels",$
        xtitle="!6Number of electrons !8n!de!n!6 in a pixel",ytitle="!6Total number of exposed traps !8n!dq!n!6 between a pixel and the readout register"
      if not keyword_set(noplottheory) then begin
        oplot,replicate(bf_notch_depth,2),yrange,psym=-3,linestyle=5 ; Notch depth line
        xyouts,orientation=90,bf_notch_depth*1.1,yrange[1]*0.9,align=1,"!6(Fitted) notch depth !8d!6"
      endif
      for di=0,trails.n_date_bins-1 do begin
        for bi=0,trails.n_back_bins-1 do begin
          if not keyword_set(postscript) then color=255+(bi-trails.n_back_bins+1)*240./trails.n_back_bins
          for ti=0,trails.n_tran_bins-1 do begin
            n_xx=50
            xx=bf_notch_depth*1.01-1+((xrange[1]-bf_notch_depth+1)^(1./(n_xx-1)))^findgen(n_xx)
            yy=acs_cte_model([[replicate(mean(trails.tran[ti,*,di,bi]),n_xx)],[xx],[replicate(mean(trails.date[ti,*,di,bi]),n_xx)],[replicate(mean(trails.back[ti,*,di,bi]),n_xx)]],params);*4
            if not keyword_set(noplottheory) then begin
              oplot,[-bf_full_well_depth,bf_notch_depth],[0,0],psym=-3,linestyle=1,color=color ; Theoretical model
              oplot,xx,yrange[0]>yy<yrange[1],psym=-3,linestyle=1,color=color ; Theoretical model
              ;oplot,xx,yrange[0]>(yy*(params[0]+perror[0])/params[0])<yrange[1],psym=-3,linestyle=1,color=color ; Theoretical model
              ;oplot,xx,yrange[0]>(yy*(params[0]-perror[0])/params[0])<yrange[1],psym=-3,linestyle=1,color=color ; Theoretical model
            endif
            oplot,trails.flux[ti,*,di,bi],y[ti,*,di,bi],psym=-3,color=color,thick=bi+1 ; Line joining data points
            for fi=0,trails.n_flux_bins-1 do begin
              usersym,[-0.5,0.5],[0,0] ; Set psym=8 to be a horizontal line
              if finite(y_error[ti,fi,di,bi]) then oplot,replicate(trails.flux[ti,fi,di,bi],2),yrange[0]>(y[ti,fi,di,bi]+[-1,1]*y_error[ti,fi,di,bi])<yrange[1],psym=-8,color=color ; Vertical error bars
              usersym,[0,0],[-0.5,0.5] ; Set psym=8 to be a vertical line
              oplot,xrange[0]>(trails.flux[ti,fi,di,bi]+[-1,1]*trails.flux_rms[ti,fi,di,bi])<xrange[1],replicate(y[ti,fi,di,bi],2),psym=-8,color=color ; Horizontal error bars
            endfor
          endfor
        endfor
      endfor
      if keyword_set(postscript) then cps
    endif
  endfor
  if ii gt 0 then begin
    ; Fix notch depth
    parinfo[5].value=max(trails.metaparameters[*,5],bestfit) & parinfo[5].fixed=1B
    parinfo[7].value=trails.metaparameters[bestfit,7] & parinfo[7].fixed=1B
    print,"Fixing notch depth at "+strtrim(parinfo[5].value,2)+" and well filling power at "+strtrim(parinfo[7].value,2)
  endif
endfor

; Finish off plot
if keyword_set(plots) then begin
  !p.multi=0
  !x.margin=xmargin
  !y.margin=ymargin
endif

end

;************************************************************

pro acs_measure_cte, PIXELS, TRAILS, $
                     DATA_DIR=data_dir, $
                     SUFFIX=suffix, $
                     EXTENSION=extension, $
                     AMPLIFIERS=amplifiers, $
                     N_SPECIES=n_species, $
                     TRAIL_LENGTHS=trail_lengths, $
                     FIXED_TRAP_RATIO=fixed_trap_ratio, $
                     ADDITIVE_CONSTANT=additive_constant, $
                     POSITIVE_DENSITIES=positive_densities, $
                     PARAMS=params, $
                     INDEX=index, $
                     SPLIT=split, $
                     FIND=find, $
                     N_TRAN_BINS=n_tran_bins, $
                     N_FLUX_BINS=n_flux_bins, $
                     N_DATE_BINS=n_date_bins, $
                     N_BACK_BINS=n_back_bins, $
                     TRAN_POWER=tran_power, $
                     FLUX_POWER=flux_power, $
                     DATE_POWER=date_power, $
                     BACK_POWER=back_power, $
                     NOTCH_DEPTH=notch_depth, $
                     WELL_FILL_POWER=well_fill_power, $
                     N_VARIABLES=n_variables, $
                     LINEAR_ALGEBRA=linear_algebra, $
                     SKIPPY=skippy, $
                     NOPLOTTHEORY=noplottheory, $
                     PLOTS=plots, $
                     POSTSCRIPT=postscript

;+
; NAME:
;      ACS_MEASURE_CTE
;
; CATEGORY:
;      Pixel-by-pixel correction of trailing due to charge transfer inefficiency in CCD detectors.
;
; PURPOSE:
;      Detect warm pixels in an image and measure the trails around them.
;
; INPUTS:
;      
;
; OPTIONAL INPUTS:
;      PIXELS - A mid-reduction data product, to enable prior steps to be skipped.
;      TRAILS - A mid-reduction data product, to enable prior steps to be skipped.
;      SUFFIX - String containing the suffix (if any) used to identify the filenames of interest.
;               For HST ACS data, this is "_raw". Alternatively, this can be definted as a two-element
;               string array. In this case, the first is used to identify the images used for CTE
;               measurement, but the second to identify warm pixels (e.g. ["_cte","_raw"]).
;
; KEYWORD PARAMETERS:
;      FIND - Look at the images from scratch to locate the hot pixels.
;             Must be done at least on the first occasion this routine is run.
;
; OUTPUTS:
;      
;
; MODIFICATION HISTORY:
;      Jan 08 - Written by Richard Massey.
;-

; Parse inputs and defaults
if not keyword_set(suffix)     then suffix="_raw" else if suffix[0] eq "_cte" then suffix=["_cte","_raw"] 
if not keyword_set(extension)  then extension="fits"
if not keyword_set(amplifiers) then amplifiers=["A","B","C","D"]


; Locate relevant ACS data (after this, the "files" variable contains neither the suffix nor the extension)
files=acs_find_data(index=index,data_dir=data_dir,suffix=suffix[0],extension=extension,dates=file_dates,/check_bias)
print,string((max(file_dates)+min(file_dates))/2.-2452334.5)+string((max(file_dates)-min(file_dates))/2.)+" "+data_dir
;return


; Split regions of the CCD read through separate amplifiers into separate files (after this, the "split_files" variable contains the suffix but not the extension)
if keyword_set(split) then acs_split, strcompress(files+suffix[0]+"."+extension,/REMOVE_ALL), /blev_linear, stats=stats
if keyword_set(stats) then save,stats,filename=data_dir+"stats.sav"
split_files=strcompress(files[indgen(n_elements(amplifiers)*n_elements(files))/n_elements(amplifiers)]+suffix[0]+amplifiers[indgen(n_elements(amplifiers)*n_elements(files)) mod n_elements(amplifiers)],/REMOVE_ALL)


; Identify warm pixels on each image and store the surrounding data
if keyword_set(find) then begin
  if n_elements(suffix) gt 1 then begin
    pixel_files=strcompress(files[indgen(n_elements(amplifiers)*n_elements(files))/n_elements(amplifiers)]+suffix[1]+amplifiers[indgen(n_elements(amplifiers)*n_elements(files)) mod n_elements(amplifiers)],/REMOVE_ALL)
    ;message,"Is this really refinding the warm pixels?"
    ;if not file_test() then 
     ;acs_find_warm_pixels, pixel_files
  endif
  acs_find_warm_pixels, split_files, PIXEL_FILES=pixel_files, BACKGROUNDS=backgrounds;, PLOTS=plots
  if file_test(data_dir+"stats.sav") then begin 
    restore,data_dir+"stats.sav" & stats.background=reform(backgrounds,n_elements(amplifiers),n_elements(files)) & save,stats,filename=data_dir+"stats.sav"
    spawn,"touch acs_stats_data.txt"
    spawn,"echo "+string((max(stats.date)+min(stats.date))/2.)+string((max(stats.date)-min(stats.date))/2.)+string(n_elements(files))+" "+stats(0).pi+string(stats(0).pid)+" "+$
          " gain"+string(mean(stats.gain))+string(max(stats.gain)-min(stats.gain))+$
                  string(mean(stats.biasgain))+string(max(stats.biasgain)-min(stats.biasgain))+$
          " back"+string(mean(stats.background))+string(stddev(stats.background))+string(stddev(stats.background)/sqrt(n_elements(files)))+string(max(stats.background)-min(stats.background))+$
                  string(mean(stats.background[0,*]))+string(stddev(stats.background[0,*])/sqrt(n_elements(files)))+string(max(stats.background[0,*])-min(stats.background[0,*]))+$
                  string(mean(stats.background[1,*]))+string(stddev(stats.background[1,*])/sqrt(n_elements(files)))+string(max(stats.background[1,*])-min(stats.background[1,*]))+$
                  string(mean(stats.background[2,*]))+string(stddev(stats.background[2,*])/sqrt(n_elements(files)))+string(max(stats.background[2,*])-min(stats.background[2,*]))+$
                  string(mean(stats.background[3,*]))+string(stddev(stats.background[3,*])/sqrt(n_elements(files)))+string(max(stats.background[3,*])-min(stats.background[3,*]))+$
          " bias"+string(mean(stats.bias))+string(stddev(stats.bias))+string(max(stats.bias)-min(stats.bias))+$
                  string(mean(stats.blev[*,0,*]))+string(stddev(stats.blev[*,0,*]))+string(max(stats.blev[*,0,*])-min(stats.blev[*,0,*]))+$ ; blev across all four amps, and all images
                  string(mean(stats.blev[*,1,*]))+string(stddev(stats.blev[*,1,*]))+string(max(stats.blev[*,1,*])-min(stats.blev[*,1,*]))+$ ; blev gradients
          " fits"+string(mean(stats.bscale))+string(max(stats.bscale)-min(stats.bscale))+$
                  string(mean(stats.bzero))+string(max(stats.bzero)-min(stats.bzero))+$
                  string(mean(stats.ltv1))+string(max(stats.ltv1)-min(stats.ltv1))+" "+stats(0).bunit+" "+data_dir+" "+systime()+" >> acs_stats_data.txt"
  endif
  delvarx,pixels
endif
;return


; Read in the data from around warm pixels and divide it up usefully
if not keyword_set(trails) then begin
  if not keyword_set(pixels) then pixels=acs_read_warm_pixels(split_files,split=3,discard=3);,name=file_basename(data_dir));strmid(file_dirname(data_dir),strpos(file_dirname(data_dir),path_sep(),/reverse_search)+1))
  trails=acs_stack_warm_pixels(pixels,n_tran_bins=n_tran_bins,n_flux_bins=n_flux_bins,n_date_bins=n_date_bins,n_back_bins=n_back_bins,warm_if_common=(n_elements(files)*0.9)>1);,tran_bins_low=1633,tran_bins_high=2038)
endif
if tag_exist(trails,"TRAIL_AMPLITUDES") and keyword_set(skippy) then goto,skippy_place


; Strip unwanted tags from trails structure
tagnames=tag_names(trails)
keep_tags=where(strmatch(tagnames,"TRAIL_LENGTHS",/FOLD_CASE)+strmatch(tagnames,"TRAIL_LENGTHS_ERROR",/FOLD_CASE)+$
                strmatch(tagnames,"TRAIL_AMPLITUDES",/FOLD_CASE)+strmatch(tagnames,"TRAIL_AMPLITUDES_ERROR",/FOLD_CASE)+$
                strmatch(tagnames,"ADDITIVE_CONSTANTS",/FOLD_CASE)+strmatch(tagnames,"ADDITIVE_CONSTANTS_ERROR",/FOLD_CASE)+$
                strmatch(tagnames,"METAPARAMETERS",/FOLD_CASE)+strmatch(tagnames,"METAPARAMETERS_ERROR",/FOLD_CASE) eq 0,n_keep_tags)
if n_keep_tags ne n_tags(trails) then begin
  new_trails=create_struct(tagnames[keep_tags[0]],trails.(keep_tags[0]))
  for i=1,n_keep_tags-1 do new_trails=create_struct(temporary(new_trails),tagnames[keep_tags[i]],trails.(keep_tags[i]))
  trails=temporary(new_trails)
endif
  

; Fit trail lengths and amplitudes in bright/far bin
if keyword_set(fixed_trap_ratio) then if n_elements(fixed_trap_ratio) gt 1 then trap_ratio=fixed_trap_ratio
if not keyword_set(trail_lengths) or not keyword_set(trap_ratio) then begin
  message,/INFO,"Fitting lengths of CTE trails using the brightest pixels furthest from the readout register."
  ti=trails.n_tran_bins-1
  fi=trails.n_flux_bins-1
  di=trails.n_date_bins-1
  bi=trails.n_back_bins-1
  if keyword_set(plots) and keyword_set(postscript) then ops,"acs_fit_cte_trail.eps",/encap,xsize=20,ysize=18
  xmargin=!x.margin & ymargin=!y.margin
  !x.margin=[6,0.25] & !y.margin=[3.25,0.25]
  trail=acs_fit_cte_trail(trail_lengths=trail_lengths,trap_ratio=trap_ratio,trails.trails[*,ti,fi,di,bi],trails.trails_error[*,ti,fi,di,bi],plots=plots,additive_constant=additive_constant,positive_densities=positive_densities)
  !x.margin=xmargin & !y.margin=ymargin
  if keyword_set(plots) and keyword_set(postscript) then cps
  trail_lengths=trail.trail_lengths
  trail_lengths_error=trail.trail_lengths_error
  pause=1B
endif else trail_lengths_error=replicate(0,n_elements(trail_lengths))
n_species=n_elements(trail_lengths)


; Append data (reinsert tags) to global trails structure
trails=create_struct(trails,"TRAIL_LENGTHS",trail_lengths[sort(trail_lengths)],$
                            "TRAIL_LENGTHS_ERROR",trail_lengths_error[sort(trail_lengths)],$
                            "TRAIL_AMPLITUDES",fltarr(n_species,trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins), $
                            "TRAIL_AMPLITUDES_ERROR",fltarr(n_species,trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins),$
                            "ADDITIVE_CONSTANTS",fltarr(n_species,trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins), $
                            "ADDITIVE_CONSTANTS_ERROR",fltarr(n_species,trails.n_tran_bins,trails.n_flux_bins,trails.n_date_bins,trails.n_back_bins))


; Summarise results so far
if keyword_set(fixed_trap_ratio) then begin
  if not keyword_set(trap_ratio) then trap_ratio=trail.trail_amplitudes[*]/trail.trail_amplitudes[0]
  messagestring=", with fixed relative density "+strtrim(trap_ratio,2)
endif else messagestring=replicate("",n_species)
for i=0,n_species-1 do message,/INFO,"Using trails of length "+strtrim(trails.trail_lengths[i],2)+"+/-"+strtrim(trails.trail_lengths_error[i],2)+" pixels"+messagestring[i]+"."
if keyword_set(pause) or keyword_set(plots) and not keyword_set(postscript) then read,systime()+": ",kk


; Fit trail amplitudes in each bin
acs_fit_cte_trails, trails, trap_ratio=trap_ratio, additive_constant=additive_constant, positive_densities=positive_densities, plots=plots, postscript=keyword_set(postscript)?file_basename(data_dir):0B
if keyword_set(plots) and not keyword_set(postscript) then read,systime()+": ",kk
skippy_place:


; Fit CTE metamodel (First set up default guesses and limits)
acs_fit_cte_model, trails, fixed_trap_ratio=fixed_trap_ratio, notch_depth=notch_depth, well_fill_power=well_fill_power, linear_algebra=linear_algebra, noplottheory=noplottheory, plots=plots, postscript=keyword_set(postscript)?file_basename(data_dir):0B


; Report to the world
if file_test(data_dir+"stats.sav") then restore,data_dir+"stats.sav"
spawn,"touch acs_cte_output_data.txt"
if keyword_set(stats) then begin
  spawn,"echo "+string((max(stats.date)+min(stats.date))/2.)+string((max(stats.date)-min(stats.date))/2.)+string(n_elements(files))+" "+$
                string(total(trails.metaparameters[*,0]))+string(total(trails.metaparameters_error[*,0]))+$
             " "+stats(0).pi+strjoin(replicate(" ",(20-strlen(stats(0).pi))>1))+string(stats(0).pid)+string(n_elements(files))+" "+$
        " gain"+string(mean(stats.gain))+string(max(stats.gain)-min(stats.gain))+$
                string(mean(stats.biasgain))+string(max(stats.biasgain)-min(stats.biasgain))+$
        " back"+string(mean(stats.background))+string(stddev(stats.background))+string(stddev(stats.background)/sqrt(n_elements(files)))+string(max(stats.background)-min(stats.background))+$
                string(mean(stats.background[0,*]))+string(stddev(stats.background[0,*])/sqrt(n_elements(files)))+string(max(stats.background[0,*])-min(stats.background[0,*]))+$
                string(mean(stats.background[1,*]))+string(stddev(stats.background[1,*])/sqrt(n_elements(files)))+string(max(stats.background[1,*])-min(stats.background[1,*]))+$
                string(mean(stats.background[2,*]))+string(stddev(stats.background[2,*])/sqrt(n_elements(files)))+string(max(stats.background[2,*])-min(stats.background[2,*]))+$
                string(mean(stats.background[3,*]))+string(stddev(stats.background[3,*])/sqrt(n_elements(files)))+string(max(stats.background[3,*])-min(stats.background[3,*]))+$
        " bias"+string(mean(stats.bias))+string(stddev(stats.bias))+string(max(stats.bias)-min(stats.bias))+$
                string(mean(stats.blev[*,0,*]))+string(stddev(stats.blev[*,0,*]))+string(max(stats.blev[*,0,*])-min(stats.blev[*,0,*]))+$ ; blev across all four amps, and all images
                string(mean(stats.blev[*,1,*]))+string(stddev(stats.blev[*,1,*]))+string(max(stats.blev[*,1,*])-min(stats.blev[*,1,*]))+$ ; blev gradients
        " fits"+string(mean(stats.bscale))+string(max(stats.bscale)-min(stats.bscale))+$
                string(mean(stats.bzero))+string(max(stats.bzero)-min(stats.bzero))+$
                string(mean(stats.ltv1))+string(max(stats.ltv1)-min(stats.ltv1))+" "+stats(0).bunit+" "+data_dir+" "+systime()+" >> acs_cte_output_data.txt"
endif else begin
  spawn,"echo "+string((max(file_dates)+min(file_dates))/2.-2452334.5)+string((max(file_dates)-min(file_dates))/2.)+string(total(trails.metaparameters[*,0]))+string(total(trails.metaparameters_error[*,0]))+$
                string(total(trails.metaparameters[*,0]))+string(total(trails.metaparameters_error[*,0]))+" "+data_dir+" "+systime()+" >> acs_cte_output_data.txt"
endelse


end
