function acs_stack_warm_pixels, pixels, $
                                n_flux_bins=n_flux_bins, $
                                n_tran_bins=n_tran_bins, $
                                n_date_bins=n_date_bins, $
                                n_back_bins=n_back_bins, $
                                warm_if_common=warm_if_common, $
                                tran_bins_low=tran_bins_low, $
                                tran_bins_high=tran_bins_high

;+
; NAME:
;      CTE_MEASURE_TRAILS
;
; CATEGORY:
;      Pixel-by-pixel correction of trailing due to charge transfer inefficiency in CCD detectors.
;
; PURPOSE:
;      Measure trailing.
;
; INPUTS:
;      PIXELS - IDL structure containing pixel locations, fluxes and surrounding pixel values, etc.
;
; OPTIONAL INPUTS:
;      None.
;
; KEYWORD PARAMETERS:
;      None.
;
; OUTPUTS:
;      PIXELS - Returns a structure containing all the necessary information.
;
; MODIFICATION HISTORY:
;      Feb 10 - Error catching when there are no pixels in a bin by RM.
;      Jan 08 - Written by Richard Massey.
;-

full_well_depth=84700
n_pixels=(size(pixels.leading))[1]
if not keyword_set(warm_if_common) then warm_if_common=10

; Decide number of bins in each variable, and introduce slight gaps between the bins and specify them differently
message,/INFO,"Deciding ranges of position/flux/date/background bins."
if not keyword_set(tran_bins_low) or not keyword_set(tran_bins_high) then begin
  if not keyword_set(n_tran_bins) then n_tran_bins=5
  tran_bins=n_pixels+findgen(n_tran_bins+1)*(2048-2*n_pixels)/float(n_tran_bins)
  tran_bins_low =fix(tran_bins[0:n_tran_bins-1])
  tran_bins_high=fix(tran_bins[1:n_tran_bins])-1
endif else n_tran_bins=n_elements(tran_bins_low)
if not keyword_set(flux_bins_low) or not keyword_set(flux_bins_high) then begin
  if not keyword_set(n_flux_bins) then n_flux_bins=10
;  max_flux=300
;  flux_bins=40*(10^(alog10((max_flux+10)/10.)/(n_flux_bins)))^indgen(n_flux_bins+1)+2600;min(pixels.flux)
  max_flux=300
  flux_bins=100*(10^(alog10((max_flux+10)/1.)/(n_flux_bins)))^indgen(n_flux_bins+1)
  flux_bins_low =flux_bins[0:n_flux_bins-1]
  flux_bins_high=flux_bins[1:n_flux_bins]-0.01 & flux_bins_high[n_flux_bins-1]=!values.f_infinity
endif else n_flux_bins=n_elements(flux_bins_low)
flux_bins_high=flux_bins_high<(2.^16*0.9);flux_bins_high<(full_well_depth*0.9)
flux_bins_low=flux_bins_low<flux_bins_high
if not keyword_set(date_bins_low) or not keyword_set(date_bins_high) then begin
  if not keyword_set(n_date_bins) then n_date_bins=1
  dates=pixels[uniq(pixels.date)].date & dates=dates[sort(dates)] & dates=dates[uniq(dates)] & date_range=[min(dates,max=max),max]
  date_bins=date_range[0]-0.1+findgen(n_date_bins+1)*(date_range[1]-date_range[0]+0.3)/float(n_date_bins)
  date_bins_low =date_bins[0:n_date_bins-1]
  date_bins_high=date_bins[1:n_date_bins]-0.1
endif else n_date_bins=n_elements(date_bins_low)
if not keyword_set(back_bins_low) or not keyword_set(back_bins_high) then begin
  if not keyword_set(n_back_bins) then n_back_bins=7
  back_range=[min(pixels.background,max=max),max]
  back_bins=back_range[0]-.1+findgen(n_back_bins+1)*(back_range[1]-back_range[0]+0.3)/float(n_back_bins)
  back_bins_low =back_bins[0:n_back_bins-1]
  back_bins_high=back_bins[1:n_back_bins]-0.1
endif else n_back_bins=n_elements(back_bins_low)
print,tran_bins_low,tran_bins_high
print,flux_bins_low,flux_bins_high
print,date_bins_low,date_bins_high
print,back_bins_low,back_bins_high


; Loop over each flux/y/date/background level bin
message,/INFO,"Sorting CTE trails into flux bins."
whichflux=ulonarr(n_elements(pixels)+1,n_flux_bins)
for fi=0,n_flux_bins-1 do begin
  in_bin=where(abs(pixels.flux      -(flux_bins_high[fi]+flux_bins_low[fi])/2.) le (flux_bins_high[fi]-flux_bins_low[fi])/2.,n_in_bin)
  whichflux[0,fi]=n_in_bin
  print,whichflux[0,fi]
  if n_in_bin gt 0 then whichflux[1:n_in_bin,fi]=in_bin
endfor
whichflux=whichflux[0:max(whichflux[0,*]),*]
;help,whichflux

; Remove pixels that are not warm on many successive exposures, to remove read noise peaks
message,/INFO,"Removing pixels that are not consistently warm."
pixel_locations=bytarr(max(pixels.x)+1,max(pixels.y)+1,n_flux_bins)
for fi=0,n_flux_bins-1 do begin
  if whichflux[0,fi] gt 0 then $
    for i=0ULL,whichflux[0,fi]-1 do pixel_locations[pixels[whichflux[i,fi]].x,pixels[whichflux[i,fi]].y,0>(fi-1):(fi+1)<(n_flux_bins-1)]+=1B
  ;for i=0ULL,whichflux[0,fi]-1 do pixel_locations[pixels[whichflux[i,fi]].x,pixels[whichflux[i,fi]].y,fi]+=1B
endfor
for fi=0,n_flux_bins-1 do begin
  good=bytarr(whichflux[0,fi]+1)
  for i=1ULL,whichflux[0,fi] do good[i]=pixel_locations[pixels[whichflux[i,fi]].x,pixels[whichflux[i,fi]].y,fi] ge warm_if_common
  good = where(good,n_good)
  print,whichflux[0,fi]," ->",n_good," ("+strtrim(100.*n_good/whichflux[0,fi],2)+"%)"
  whichflux[0,fi]=n_good
  if n_good gt 0 then whichflux[1:n_good,fi]=whichflux[good,fi]
  if n_good lt (size(whichflux,/DIMENSIONS))[1]-1 then whichflux[n_good+1,fi]=0
endfor
delvarx,pixel_locations,good

message,/INFO,"Sorting CTE trails into position bins."
whichtran=ulonarr(n_elements(pixels)+1,n_tran_bins)
for ti=0,n_tran_bins-1 do begin
  in_bin=where(abs(pixels.y         -(tran_bins_high[ti]+tran_bins_low[ti])/2.) le (tran_bins_high[ti]-tran_bins_low[ti])/2.,n_in_bin)
  whichtran[0,ti]=n_in_bin
  print,whichtran[0,ti]
  if n_in_bin gt 0 then whichtran[1:n_in_bin,ti]=in_bin
endfor
whichtran=whichtran[0:max(whichtran[0,*]),*]
some_in_bin=where(whichtran[0,*] gt 0,n_tran_bins)
whichtran=whichtran[*,some_in_bin]
date_bins_high=date_bins_high[some_in_bin]
date_bins_low=date_bins_low[some_in_bin]
;help,whichtran

message,/INFO,"Sorting CTE trails into date bins."
whichdate=ulonarr(n_elements(pixels)+1,n_date_bins)
for di=0,n_date_bins-1 do begin
  in_bin=where(abs(pixels.date      -(date_bins_high[di]+date_bins_low[di])/2.) le (date_bins_high[di]-date_bins_low[di])/2.,n_in_bin)
  whichdate[0,di]=n_in_bin
  print,whichdate[0,di]
  if n_in_bin gt 0 then whichdate[1:n_in_bin,di]=in_bin
endfor
whichdate=whichdate[0:max(whichdate[0,*]),*]
some_in_bin=where(whichdate[0,*] gt 0,n_date_bins)
whichdate=whichdate[*,some_in_bin]
date_bins_high=date_bins_high[some_in_bin]
date_bins_low=date_bins_low[some_in_bin]
;help,whichdate

message,/INFO,"Sorting CTE trails into background bins."
whichback=ulonarr(n_elements(pixels)+1,n_back_bins)
for bi=0,n_back_bins-1 do begin
  in_bin=where(abs(pixels.background-(back_bins_high[bi]+back_bins_low[bi])/2.) le (back_bins_high[bi]-back_bins_low[bi])/2.,n_in_bin)
  whichback[0,bi]=n_in_bin
  print,whichback[0,bi]
  if n_in_bin gt 0 then whichback[1:n_in_bin,bi]=in_bin
endfor
whichback=whichback[0:max(whichback[0,*]),*]
some_in_bin=where(whichback[0,*] gt 0,n_back_bins)
whichback=whichback[*,some_in_bin]
back_bins_high=back_bins_high[some_in_bin]
back_bins_low=back_bins_low[some_in_bin]
;help,whichback


; Initialise variables to contain the results
trails=fltarr(n_pixels,n_tran_bins,n_flux_bins,n_date_bins,n_back_bins)       & trails[*]=!values.f_nan
trails_error=fltarr(n_pixels,n_tran_bins,n_flux_bins,n_date_bins,n_back_bins) & trails_error[*]=!values.f_infinity
n_trails=lonarr(n_tran_bins,n_flux_bins,n_date_bins,n_back_bins)              & n_trails[*]=0
tran=fltarr(n_tran_bins,n_flux_bins,n_date_bins,n_back_bins)                  & tran[*]=0;!values.f_nan
flux=fltarr(n_tran_bins,n_flux_bins,n_date_bins,n_back_bins)                  & flux[*]=0;!values.f_nan
date=fltarr(n_tran_bins,n_flux_bins,n_date_bins,n_back_bins)                  & date[*]=0;!values.f_nan
back=fltarr(n_tran_bins,n_flux_bins,n_date_bins,n_back_bins)                  & back[*]=0;!values.f_nan
tran_rms=fltarr(n_tran_bins,n_flux_bins,n_date_bins,n_back_bins)              & tran_rms[*]=!values.f_infinity
flux_rms=fltarr(n_tran_bins,n_flux_bins,n_date_bins,n_back_bins)              & flux_rms[*]=!values.f_infinity
date_rms=fltarr(n_tran_bins,n_flux_bins,n_date_bins,n_back_bins)              & date_rms[*]=!values.f_infinity
back_rms=fltarr(n_tran_bins,n_flux_bins,n_date_bins,n_back_bins)              & back_rms[*]=!values.f_infinity

; Loop over each flux/y/date/background level bin
message,/INFO,"Averaging CTE trail within each position/flux/date/background bin."
for fi=0,n_flux_bins-1 do begin
  for ti=0,n_tran_bins-1 do begin
    ; Progress report
    print,fix(ti*100./n_tran_bins/n_flux_bins+fi*100./n_flux_bins),"%...",format="(I0,A0,$)"
    for di=0,n_date_bins-1 do begin
      for bi=0,n_back_bins-1 do begin
            ; Default values
            tran[ti,fi,di,bi]=(tran_bins_high[di]+tran_bins_low[di])/2.
            flux[ti,fi,di,bi]=(flux_bins_high[di]+flux_bins_low[di])/2.
            date[ti,fi,di,bi]=(date_bins_high[di]+date_bins_low[di])/2.
            back[ti,fi,di,bi]=(back_bins_high[di]+back_bins_low[di])/2.
        ; Real values
        if n_date_bins gt 1 then message,"Need to uncomment the following line!"
        ;if whichflux[0,fi]+whichtran[0,ti]+whichback[0,bi]+whichflux[0,di] gt 0 then begin
        if whichflux[0,fi]+whichtran[0,ti]+whichback[0,bi] gt 0 then begin
          in_all_bin=0
          if whichflux[0,fi] gt 0 then in_all_bin=[in_all_bin,whichflux[1:whichflux[0,fi],fi]]
          if whichtran[0,ti] gt 0 then in_all_bin=[in_all_bin,whichtran[1:whichtran[0,ti],ti]]
          if whichback[0,bi] gt 0 then in_all_bin=[in_all_bin,whichback[1:whichback[0,bi],bi]]
          ;if whichflux[0,di] gt 0 then in_all_bin=[in_all_bin,whichdate[1:whichdate[0,di],di]]
          ; Cannibalised from setintersection.pro
          min_in_bin = min(in_all_bin[1:*], MAX=max_in_bin)
          in_bin=where(histogram(in_all_bin[1:*], MIN=min_in_bin, MAX=max_in_bin) ge 3,n_in_bin)+min_in_bin
          print,n_in_bin;,format="(I0,$)"
          if n_in_bin gt 1 then begin
            ;print,in_bin[0:5<n_in_bin]
            n_trails[ti,fi,di,bi]=n_in_bin
            ; Evaluate quantities of interest in each bin ; Subtract off very faint trail above warm pixel?
            tran[ti,fi,di,bi]=mean(pixels[in_bin].y)
            flux[ti,fi,di,bi]=mean(pixels[in_bin].flux)
            date[ti,fi,di,bi]=mean(pixels[in_bin].date)
            back[ti,fi,di,bi]=mean(pixels[in_bin].background)
            tran_rms[ti,fi,di,bi]=stddev(pixels[in_bin].y);/float(n_in_bin)
            flux_rms[ti,fi,di,bi]=stddev(pixels[in_bin].flux);/float(n_in_bin)
            date_rms[ti,fi,di,bi]=stddev(pixels[in_bin].date);/float(n_in_bin)
            back_rms[ti,fi,di,bi]=stddev(pixels[in_bin].background);/float(n_in_bin)
            blah=float(reform(pixels[in_bin].trailing)-reform(pixels[in_bin].leading))
            trails[*,ti,fi,di,bi]=total(blah,2)/float(n_in_bin)
            trails_error[*,ti,fi,di,bi]=sqrt((total(blah^2,2)-total(blah,2)^2/float(n_in_bin))/(float(n_in_bin)-1))/sqrt(n_in_bin)
            ;print,trails[*,ti,fi,di,bi]
          endif
        endif
      endfor
    endfor
  endfor
endfor
print,100,"%.",format="(I0,A0)"

;; Loop over each flux/y/date/background level bin
;message,/INFO,"Averaging CTE trail within each position/flux/date/background bin."
;for fi=0,n_flux_bins-1 do begin
;  for ti=0,n_tran_bins-1 do begin
;    ; Progress report
;    print,fix(ti*100./n_tran_bins/n_flux_bins+fi*100./n_flux_bins),"%...",format="(I0,A0,$)"
;    for di=0,n_date_bins-1 do begin
;      for bi=0,n_back_bins-1 do begin
;        ; Identify which pixels lie in this region of parameter space
;        ;in_bin=where(pixels.flux       ge flux_bins_low[fi]  and $
;        ;             pixels.flux       lt flux_bins_high[fi] and $
;        ;             pixels.y          ge tran_bins_low[ti]  and $
;        ;             pixels.y          lt tran_bins_high[ti] and $
;        ;             pixels.date       ge date_bins_low[di]  and $
;        ;             pixels.date       lt date_bins_high[di] and $
;        ;             pixels.background ge back_bins_low[bi]  and $
;        ;             pixels.background lt back_bins_high[bi], n_in_bin)
;        in_bin=where(abs(pixels.flux      -(flux_bins_high[fi]+flux_bins_low[fi])/2.) le (flux_bins_high[fi]-flux_bins_low[fi])/2. and $
;                     abs(pixels.y         -(tran_bins_high[ti]+tran_bins_low[ti])/2.) le (tran_bins_high[ti]-tran_bins_low[ti])/2. and $
;                     abs(pixels.date      -(date_bins_high[di]+date_bins_low[di])/2.) le (date_bins_high[di]-date_bins_low[di])/2. and $
;                     abs(pixels.background-(back_bins_high[bi]+back_bins_low[bi])/2.) le (back_bins_high[bi]-back_bins_low[bi])/2., n_in_bin)
;        n_trails[ti,fi,di,bi]=n_in_bin
;        if n_in_bin gt 0 then begin
;          ; Evaluate quantities of interest in each bin ; Subtract off very faint trail above warm pixel?
;          tran[ti,fi,di,bi]=mean(pixels[in_bin].y)
;          flux[ti,fi,di,bi]=mean(pixels[in_bin].flux)
;          date[ti,fi,di,bi]=mean(pixels[in_bin].date)
;          back[ti,fi,di,bi]=mean(pixels[in_bin].background)
;          tran_rms[ti,fi,di,bi]=stddev(pixels[in_bin].y);/float(n_in_bin)
;          flux_rms[ti,fi,di,bi]=stddev(pixels[in_bin].flux);/float(n_in_bin)
;          date_rms[ti,fi,di,bi]=stddev(pixels[in_bin].date);/float(n_in_bin)
;          back_rms[ti,fi,di,bi]=stddev(pixels[in_bin].background);/float(n_in_bin)
;          blah=float(reform(pixels[in_bin].trailing)-reform(pixels[in_bin].leading))
;          trails[*,ti,fi,di,bi]=total(blah,2)/float(n_in_bin)
;          trails_error[*,ti,fi,di,bi]=sqrt((total(blah^2,2)-total(blah,2)^2/float(n_in_bin))/(float(n_in_bin)-1))/sqrt(n_in_bin)
;;          blah=float(reform(pixels[in_bin].trailing)-reform(pixels[in_bin].leading))/((sqrt(pixels[in_bin].flux)-sqrt(pixels[in_bin].background))##replicate(1.,n_pixels))
;;          trails[*,ti,fi,di,bi]=total(blah,2)/n_in_bin * (sqrt(flux_bins_med[fi])-sqrt(back_bins_med[bi]))
;;          trails_error[*,ti,fi,di,bi]=sqrt((total(blah^2,2)-total(blah,2)^2/n_in_bin)/(n_in_bin-1))/sqrt(n_in_bin) * (sqrt(flux_bins_med[fi])-sqrt(back_bins_med[bi]))
;        endif
;      endfor
;    endfor
;  endfor
;endfor
;print,100,"%.",format="(I0,A0)"

; Report back one variable containing all these results
trails={trails:trails,$
        n_trails:long(n_trails),$
        trails_error:trails_error,$
        full_well_depth:full_well_depth,$
        n_pixels:n_pixels,$
        n_flux_bins:n_flux_bins,$
        n_tran_bins:n_tran_bins,$
        n_date_bins:n_date_bins,$
        n_back_bins:n_back_bins,$
        flux:flux,$
        tran:tran,$
        date:date,$
        back:back,$
        tran_rms:tran_rms,$
        flux_rms:flux_rms,$
        date_rms:date_rms,$
        back_rms:back_rms,$
        flux_bins_low:flux_bins_low,flux_bins_high:flux_bins_high,$
        tran_bins_low:tran_bins_low,tran_bins_high:tran_bins_high,$
        date_bins_low:date_bins_low,date_bins_high:date_bins_high,$
        back_bins_low:back_bins_low,back_bins_high:back_bins_high}
return,trails


end
